
__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import tempfile
import tarfile

import radical.utils as ru

from ...   import states    as rps
from ...   import constants as rpc
from ...   import utils     as rpu

from .base import TMGRStagingInputComponent

from ...staging_directives import complete_url, expand_staging_directives


# if we receive more than a certain numnber of tasks in a bulk, we create the
# task sandboxes in a remote bulk op.  That limit is defined here, along with
# the definition of the bulk mechanism used to create the sandboxes:
#   tar : unpack a locally created tar which contains all sandboxes

TASK_BULK_MKDIR_THRESHOLD = 1024 * 1024
TASK_BULK_MKDIR_MECHANISM = 'tar'


# ------------------------------------------------------------------------------
#
class Default(TMGRStagingInputComponent):
    '''
    This component performs all tmgr side input staging directives for compute
    tasks.  It gets tasks from the tmgr_staging_input_queue, in
    TMGR_STAGING_INPUT_PENDING state, will advance them to TMGR_STAGING_INPUT
    state while performing the staging, and then moves then to the
    AGENT_SCHEDULING_PENDING state, passing control to the agent.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        TMGRStagingInputComponent.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        self._pilots       = dict()
        self._pilots_lock  = ru.RLock()
        self._connected    = list()  # list of pilot conected by ZMQ
        self._session_sbox = self._reg['cfg.session_sandbox']
        self._stager       = rpu.StagingHelper(self._log)
        self._tar_idx      = 0

        self.register_input(rps.TMGR_STAGING_INPUT_PENDING,
                            rpc.TMGR_STAGING_INPUT_QUEUE, self.work)

        # this queue is inaccessible, needs routing via mongodb
        self.register_output(rps.AGENT_STAGING_INPUT_PENDING,
                             rpc.PROXY_TASK_QUEUE)

        self._mkdir_threshold = self.cfg.get('task_bulk_mkdir_threshold',
                                             TASK_BULK_MKDIR_THRESHOLD)


    # --------------------------------------------------------------------------
    #
    def control_cb(self, topic, msg):

        # keep track of `add_pilots` commands and updates self._pilots
        # accordingly.

        cmd = msg.get('cmd')
        arg = msg.get('arg')

        if cmd == 'add_pilots':

            pilots = arg['pilots']

            with self._pilots_lock:

                for pilot in pilots:
                    pid = pilot['uid']
                    self._log.debug('add pilot %s', pid)

                    if pid not in self._pilots:
                        self._pilots[pid] = pilot

        elif cmd == 'pilot_register':

            pid = arg['pid']
            self._log.debug('register pilot %s', pid)

            if pid not in self._connected:
                self._connected.append(pid)

            # let pilot know that tasks will arive via ZMQ
            self.publish(rpc.CONTROL_PUBSUB, msg={'cmd': 'pilot_register_ok',
                                                  'arg': {'pid': pid}})


    # --------------------------------------------------------------------------
    #
    def _advance_tasks(self, tasks, pid=None, state=None, push=True):

        if not state:
            state = rps.AGENT_STAGING_INPUT_PENDING

        # perform and publish state update
        # push to the proxy queue
        for task in tasks:
            self._log.debug_8('push to proxy: %s', task['uid'])

        self.advance(tasks, state, publish=True, push=push, qname=pid)


    # --------------------------------------------------------------------------
    #
    def work(self, tasks):

        self.advance(tasks, rps.TMGR_STAGING_INPUT, publish=True, push=False)

        # we first filter out any tasks which don't need any input staging, and
        # advance them again as a bulk.  We work over the others one by one, and
        # advance them individually, to avoid stalling from slow staging ops.

        session_sbox     = self._session_sbox
        staging_tasks    = dict()  # pid: [tasks]
        no_staging_tasks = dict()  # pid: [tasks]

        for task in tasks:

            pid  = task['pilot']
            if pid not in staging_tasks   : staging_tasks[pid]    = list()
            if pid not in no_staging_tasks: no_staging_tasks[pid] = list()

            # check if we have any staging directives to be enacted in this
            # component
            actionables = list()
            for sd in task['description'].get('input_staging', []):
                if sd['action'] in [rpc.TRANSFER, rpc.TARBALL]:
                    actionables.append(sd)

            if actionables:
                staging_tasks[pid].append([task, actionables])
            else:
                no_staging_tasks[pid].append(task)

        # Optimization: if we obtained a large bulk of tasks, we at this point
        # attempt a bulk mkdir for the task sandboxes, to free the agent of
        # performing that operation.  That implies that the agent needs to check
        # sandbox existence before attempting to create them now.
        #
        # Note that this relies on the tmgr scheduler to assigning the sandbox
        # to the task.
        #
        # Note further that we need to make sure that all tasks are actually
        # pointing into the same target file system, so we need to cluster by
        # filesystem before checking the bulk size.  For simplicity we actually
        # cluster by pilot ID, which is sub-optimal for task bulks which go to
        # different pilots on the same resource (think OSG).
        #
        # Note further that we skip the bulk-op for all tasks for which we
        # actually need to stage data, since the mkdir will then implicitly be
        # done anyways.
        #
        # Caveat: we can actually only (reasonably) do this if we know some
        # details about the pilot, because otherwise we'd have too much guessing
        # to do about the pilot configuration (sandbox, access schema, etc), so
        # we only attempt this optimization for tasks scheduled to pilots for
        # which we learned those details.
        sboxes = dict()  # pid: [sboxes]
        for pid in no_staging_tasks:
            for task in no_staging_tasks[pid]:
                if pid not in sboxes:
                    sboxes[pid] = list()
                sboxes[pid].append(task['task_sandbox'])

        # now trigger the bulk mkdir for all filesystems which have more than
        # a certain tasks tohandle in this bulk:
        for pid in sboxes:

            with self._pilots_lock:
                pilot = self._pilots.get(pid)

            if not pilot:
                # we don't feel inclined to optimize for unknown pilots
                self._log.debug('pid unknown - skip optimization', pid)
                continue

            task_sboxes  = sboxes[pid]

            if len(task_sboxes) >= self._mkdir_threshold:
                self._log.debug('tar %d sboxes', len(task_sboxes))

                session_sbox = self._session._get_session_sandbox(pilot)

                # create a tarball with all task sandboxes, push
                # it over, and untar it (one untar op then creates all dirs).
                if TASK_BULK_MKDIR_MECHANISM == 'tar':

                    tmp_path = tempfile.mkdtemp(prefix='rp_agent_tar_dir')
                    tmp_dir  = os.path.abspath(tmp_path)
                    tar_name = '%s.%s.%04d.tar' % (self._session.uid, self.uid,
                                                   self._tar_idx)
                    tar_tgt  = '%s/%s'         % (tmp_dir, tar_name)
                    tar_url  = ru.Url('file://localhost/%s' % tar_tgt)

                    self._tar_idx += 1

                    # we want pathnames which are relative to the session
                    # sandbox.  Ignore all other sandboxes - the agent will have
                    # to create those.
                    root = str(session_sbox)
                    rlen = len(root)
                    rels = list()
                    for path in task_sboxes:
                        if path.startswith(root):
                            rels.append(path[rlen + 1:])

                    rpu.create_tar(tar_tgt, rels)

                    tar_rem_path = "%s/%s" % (str(session_sbox), tar_name)

                    self._log.debug('sbox: %s [%s]', session_sbox,
                                                             type(session_sbox))
                    self._log.debug('copy: %s -> %s', tar_url, tar_rem_path)

                    self._prof.prof('staging_in_start', uid=pid, msg='tar')
                    self._stager.copy(tar_url, tar_rem_path)
                    self._prof.prof('staging_in_stop', uid=pid, msg='tar')

                    # get a job service handle to the target resource and run
                    # the untar command.  Use the hop to skip the batch system
                    hop_url = pilot['js_hop']
                    self._log.debug('js  : %s', hop_url)

                    cmd = "tar xvf %s/%s -C %s" % (session_sbox.path, tar_name,
                                                   session_sbox.path)

                    out, err, ret = self._stager.sh_callout(hop_url, cmd)
                    self._log.debug('untar: %s', [cmd, out, err, ret])


        for pid in no_staging_tasks:
            if no_staging_tasks[pid]:
                # nothing to stage, push to the agent
                self._advance_tasks(no_staging_tasks[pid], pid)

        to_fail = list()
        for pid in staging_tasks:
            for task, actionables in staging_tasks[pid]:
                try:
                    self._handle_task(task, actionables)
                    self._advance_tasks([task], pid)

                except Exception as e:
                    # staging failed - do not pass task to agent
                    self._log.exception('staging for %s failed', task['uid'])
                    task['control']          = 'tmgr'
                    task['exception']        = repr(e)
                    task['exception_detail'] = '\n'.join(ru.get_exception_trace())
                    to_fail.append(task)

        self._advance_tasks(to_fail, state=rps.FAILED, push=False)


    # --------------------------------------------------------------------------
    #
    def _handle_task(self, task, actionables):

        # FIXME: we should created task sandboxes in a bulk

        uid = task['uid']

        self._prof.prof("create_sandbox_start", uid=uid)

        src_context = {'pwd'      : task['client_sandbox'],     # !!!
                       'client'   : task['client_sandbox'],
                       'task'     : task['task_sandbox'],
                       'pilot'    : task['pilot_sandbox'],
                       'session'  : task['session_sandbox'],
                       'resource' : task['resource_sandbox'],
                       'endpoint' : task['endpoint_fs']}
        tgt_context = {'pwd'      : task['task_sandbox'],       # !!!
                       'client'   : task['client_sandbox'],
                       'task'     : task['task_sandbox'],
                       'pilot'    : task['pilot_sandbox'],
                       'session'  : task['session_sandbox'],
                       'resource' : task['resource_sandbox'],
                       'endpoint' : task['endpoint_fs']}

        # we have actionable staging directives, and thus we need a task
        # sandbox.
        sandbox = ru.Url(task["task_sandbox"])

        self._stager.mkdir(sandbox)
        self._prof.prof("create_sandbox_stop", uid=uid)

        # Loop over all transfer directives and filter out tarball staging
        # directives.  Those files are added into a tarball, and a single
        # actionable to stage that tarball replaces the original actionables.

        # create a new actionable list during the filtering
        new_actionables = list()
        tar_file        = None
        tar_path        = None
        tar_sd          = None

        for sd in actionables:

            # don't touch non-tar SDs
            if sd['action'] != rpc.TARBALL:
                new_actionables.append(sd)

            else:

                action = sd['action']
                did    = sd['uid']
                src    = sd['source']
                tgt    = sd['target']

                # client stager only handles remote actions
                if action not in [rpc.TRANSFER, rpc.TARBALL]:
                    self._prof.prof('staging_in_skip', uid=uid, msg=did)
                    continue

                src = complete_url(src, src_context, self._log)
                tgt = complete_url(tgt, tgt_context, self._log)

                self._prof.prof('staging_in_tar_start', uid=uid, msg=did)

                # create a tarfile on the first match, and register for transfer
                if not tar_file:
                    tmp_file = tempfile.NamedTemporaryFile(
                                                prefix='rp_usi_%s.' % uid,
                                                suffix='.tar',
                                                delete=False)
                    tar_path = tmp_file.name
                    tar_file = tarfile.open(fileobj=tmp_file, mode='w')
                    tar_src  = ru.Url('file://localhost/%s' % tar_path)
                    tar_tgt  = ru.Url('task:///%s.tar'      % uid)
                    tar_did  = ru.generate_id('sd')
                    tar_sd   = {'action' : rpc.TRANSFER,
                                'flags'  : rpc.DEFAULT_FLAGS,
                                'uid'    : tar_did,
                                'source' : str(tar_src),
                                'target' : str(tar_tgt),
                               }
                    new_actionables.append(tar_sd)

                    self._log.debug('create tar sd %s', tar_sd)

                # add the src file
                tar_file.add(src.path, arcname=tgt.path)

                self._prof.prof('staging_in_tar_stop',  uid=uid, msg=did)


        # make sure tarball is flushed to disk
        if tar_file:
            tar_file.close()

        new_actionables = expand_staging_directives(new_actionables,
                                            src_context, tgt_context, self._log)

        # work on the filtered TRANSFER actionables
        for sd in new_actionables:
            self._prof.prof('staging_in_start', uid=uid, msg=sd['uid'])
            self._stager.handle_staging_directive(sd)
            self._prof.prof('staging_in_stop', uid=uid, msg=sd['uid'])

        if tar_file:

            assert tar_path
            assert tar_sd

            # some tarball staging was done.  Add a staging directive for the
            # agent to untar the tarball, and clean up.
            tar_sd['action'] = rpc.TARBALL
            task['description']['input_staging'].append(tar_sd)
            os.remove(tar_path)


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"

import errno
import os
import shutil
import tarfile

import radical.utils as ru

from ...  import utils     as rpu
from ...  import states    as rps
from ...  import constants as rpc

from .base import AgentStagingInputComponent

from ...staging_directives import complete_url


# ------------------------------------------------------------------------------
#
class Default(AgentStagingInputComponent):
    """
    This component performs all agent side input staging directives for compute
    tasks.  It gets tasks from the agent_staging_input_queue, in
    AGENT_STAGING_INPUT_PENDING state, will advance them to AGENT_STAGING_INPUT
    state while performing the staging, and then moves then to the
    AGENT_SCHEDULING_PENDING state, into the agent_scheduling_queue.
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        AgentStagingInputComponent.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        self._pwd = os.getcwd()

        self.register_input(rps.AGENT_STAGING_INPUT_PENDING,
                            rpc.AGENT_STAGING_INPUT_QUEUE, self.work)

        self.register_output(rps.AGENT_SCHEDULING_PENDING,
                             rpc.AGENT_SCHEDULING_QUEUE)


    # --------------------------------------------------------------------------
    #
    def _work(self, tasks):

        # we first filter out any tasks which don't need any input staging, and
        # advance them again as a bulk.  We work over the others one by one, and
        # advance them individually, to avoid stalling from slow staging ops.

        no_staging_tasks = list()
        staging_tasks    = list()

        for task in ru.as_list(tasks):

            # check if we have any staging directives to be enacted in this
            # component
            actionables = list()
            for sd in task['description'].get('input_staging', []):

                if sd['action'] in [rpc.LINK, rpc.COPY, rpc.MOVE, rpc.TARBALL]:
                    actionables.append(sd)

            if actionables:
                staging_tasks.append([task, actionables])
            else:
                no_staging_tasks.append(task)

        if no_staging_tasks:
            self.advance(no_staging_tasks, rps.AGENT_SCHEDULING_PENDING,
                         publish=True, push=True)

        for task, actionables in staging_tasks:
            try:
                self._handle_task_staging(task, actionables)

            except Exception as e:
                self._log.exception('staging error')
                task['exception']        = repr(e)
                task['exception_detail'] = '\n'.join(ru.get_exception_trace())

                self.advance(task, rps.FAILED)


    # --------------------------------------------------------------------------
    #
    def _handle_task_staging(self, task, actionables):

        uid = task['uid']

        # By definition, this compoentn lives on the pilot's target resource.
        # As such, we *know* that all staging ops which would refer to the
        # resource now refer to file://localhost, and thus translate the task,
        # pilot and resource sandboxes into that scope.  Some assumptions are
        # made though:
        #
        #   * paths are directly translatable across schemas
        #   * resource level storage is in fact accessible via file://
        #
        # FIXME: URL creation and manipulation is costly and should be cached

        task_sandbox     = ru.Url(task['task_sandbox'])
        pilot_sandbox    = ru.Url(task['pilot_sandbox'])
        session_sandbox  = ru.Url(task['session_sandbox'])
        resource_sandbox = ru.Url(task['resource_sandbox'])
        endpoint_fs      = ru.Url(task['endpoint_fs'])

        task_sandbox.schema     = 'file'
        pilot_sandbox.schema    = 'file'
        session_sandbox.schema  = 'file'
        resource_sandbox.schema = 'file'
        endpoint_fs.schema      = 'file'

        task_sandbox.host       = 'localhost'
        pilot_sandbox.host      = 'localhost'
        session_sandbox.host    = 'localhost'
        resource_sandbox.host   = 'localhost'
        endpoint_fs.host        = 'localhost'

        src_context = {'pwd'      : str(task_sandbox),       # !!!
                       'task'     : str(task_sandbox),
                       'pilot'    : str(pilot_sandbox),
                       'session'  : str(session_sandbox),
                       'resource' : str(resource_sandbox),
                       'endpoint' : str(endpoint_fs)}
        tgt_context = {'pwd'      : str(task_sandbox),       # !!!
                       'task'     : str(task_sandbox),
                       'pilot'    : str(pilot_sandbox),
                       'session'  : str(session_sandbox),
                       'resource' : str(resource_sandbox),
                       'endpoint' : str(endpoint_fs)}


        # we can now handle the actionable staging directives
        for sd in actionables:

            action = sd['action']
            did    = sd['uid']
            src    = sd['source']
            tgt    = sd['target']

            self._prof.prof('staging_in_start', uid=uid, msg=did)

            # agent stager only handles local actions
            if action not in [rpc.COPY, rpc.LINK, rpc.MOVE]:
                self._prof.prof('staging_in_skip', uid=uid, msg=did)
                continue

            # Fix for when the target PATH is empty
            # we assume current directory is the task staging 'task://'
            # and we assume the file to be copied is the base filename
            # of the source
            if tgt is None: tgt = ''
            if tgt.strip() == '':
                tgt = 'task:///{}'.format(os.path.basename(src))
            # Fix for when the target PATH is exists *and* it is a folder
            # we assume the 'current directory' is the target folder
            # and we assume the file to be copied is the base filename
            # of the source
            elif os.path.exists(tgt.strip()) and os.path.isdir(tgt.strip()):
                tgt = os.path.join(tgt, os.path.basename(src))


            src = complete_url(src, src_context, self._log)
            tgt = complete_url(tgt, tgt_context, self._log)

            # Currently, we use the same schema for files and folders.
            assert tgt.schema == 'file', 'staging tgt must be file://'

            if action in [rpc.COPY, rpc.LINK, rpc.MOVE]:
                assert src.schema == 'file', 'staging src expected as file://'

            # implicitly create target dir if needed - but only for local ops
            if action != rpc.TRANSFER:
                tgtdir = os.path.dirname(tgt.path)
                if tgtdir != task_sandbox.path:
                    self._log.debug("mkdir %s", tgtdir)
                    ru.rec_makedir(tgtdir)

            if action == rpc.COPY:
                try:
                    shutil.copytree(src.path, tgt.path)
                except OSError as exc:
                    if exc.errno == errno.ENOTDIR:
                        shutil.copy(src.path, tgt.path)
                    else:
                        raise

            elif action == rpc.LINK:

                # Fix issue/1513 if link source is file and target is folder.
                # should support POSIX standard where link is created
                # with the same name as the source
                if os.path.isfile(src.path) and os.path.isdir(tgt.path):
                    os.symlink(src.path,
                               '%s/%s' % (tgt.path, os.path.basename(src.path)))

                else:
                    os.symlink(src.path, tgt.path)

            elif action == rpc.MOVE:
                shutil.move(src.path, tgt.path)

            elif action == rpc.TRANSFER:

                # NOTE:  TRANSFER directives don't arrive here right now.
                # FIXME: we only handle srm staging right now, and only for
                #        a specific target proxy. Other TRANSFER directives are
                #        left to tmgr input staging.  We should use SAGA to
                #        attempt all staging ops which do not involve the client
                #        machine.
                self._log.error('no transfer for %s -> %s', src, tgt)
                self._prof.prof('staging_in_fail', uid=uid, msg=did)
                raise NotImplementedError('unsupported transfer %s' % src)

            elif action == rpc.TARBALL:

                # If somethig was staged via the tarball method, the tarball is
                # extracted and then removed from the task folder.  The target
                # path is expected to be an *absolute* path on the target system
                # - any relative paths specified by the application are expected
                # to get expanded on the client side.
                tarball = '%s/%s.tar' % (os.path.dirname(tgt.path), uid)
                self._log.debug('extract tarball for %s', tarball)
                tar = tarfile.open(tarball)
                tar.extractall(path='/')
                tar.close()

              # FIXME: make tarball removal dependent on debug settings
              # os.remove(os.path.dirname(tgt.path) + '/' + uid + '.tar')

            self._prof.prof('staging_in_stop', uid=uid, msg=did)

        # all staging is done -- pass on to the scheduler
        self.advance(task, rps.AGENT_SCHEDULING_PENDING, publish=True, push=True)


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

# pyxlint: disable=redefined-outer-name
__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import sys
import radical.pilot as rp

# ATTENTION:
#
# This example needs significant time to run, and there is some probability that
# the larger futuregrid pilots are not getting through the batch queue at all.
# It is thus not part of the RP test suite.

# READ: The RADICAL-Pilot documentation:
#   https://radicalpilot.readthedocs.io/en/stable/
#
# Try running this example with RADICAL_PILOT_VERBOSE=debug set if
# you want to see what happens behind the scenes!


# ------------------------------------------------------------------------------
#
def pilot_state_cb (pilot, state):
    """ this callback is invoked on all pilot state changes """

    print("[Callback]: Pilot '%s' state: %s." % (pilot.uid, state))

    if state == rp.FAILED:
        sys.exit (1)


# ------------------------------------------------------------------------------
#
def task_state_cb (task, state):
    """ this callback is invoked on all task state changes """

    print("[Callback]: Task  '%s' state: %s." % (task.uid, state))

    if state == rp.FAILED:
        sys.exit (1)


# ------------------------------------------------------------------------------
#
if __name__ == "__main__":

    # we can optionally pass session name to RP
    if len(sys.argv) > 1:
        session_name = sys.argv[1]
    else:
        session_name = None

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session(uid=session_name)
    print("session id: %s" % session.uid)

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        # Add a Pilot Manager. Pilot managers manage one or more Pilots.
        pmgr = rp.PilotManager(session=session)

        # Register our callback with the PilotManager. This callback will get
        # called every time any of the pilots managed by the PilotManager
        # change their state.
        pmgr.register_callback(pilot_state_cb)

        pdesc = rp.PilotDescription()
        pdesc.resource  = "local.localhost"
        pdesc.runtime   = 12  # minutes
        pdesc.cores     = 4
        pdesc.cleanup   = True

        pilot_1 = pmgr.submit_pilots(pdesc)

        # create a second pilot with a new description
        pdesc = rp.PilotDescription()
        pdesc.resource  = "local.localhost"
        pdesc.runtime   = 40  # minutes
        pdesc.cores     = 32

        pilot_2 = pmgr.submit_pilots(pdesc)

        # reuse the pilot description for the third pilot
        pdesc.cores     = 128

        pilot_3 = pmgr.submit_pilots(pdesc)

        # Combine the Pilot, the Tasks and a scheduler via
        # a TaskManager object.
        tmgr = rp.TaskManager (session=session, scheduler=rp.SCHEDULER_BACKFILLING)

        # Register our callback with the TaskManager. This callback will get
        # called every time any of the tasks managed by the TaskManager
        # change their state.
        tmgr.register_callback(task_state_cb)

        # Add the previsouly created Pilot to the TaskManager.
        tmgr.add_pilots([pilot_1, pilot_2, pilot_3])

      # # wait until first pilots become active
        pilot_1.wait (state=rp.PMGR_ACTIVE)

        # Create a workload of 8 Tasks.
        cus = list()

        for task_count in range(0, 512):
            t = rp.TaskDescription()
            t.executable = '/bin/sleep'
            t.arguments = ['10']
            cus.append(t)

        # Submit the previously created Task descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning Tasks to the Pilots.
        tasks = tmgr.submit_tasks(cus)

        # Wait for all tasks to reach a terminal state (DONE or FAILED).
        tmgr.wait_tasks()

        pmgr.cancel_pilots ()

        for task in tasks:
            print("* Task %s state: %s, exit code: %s"
                % (task.uid, task.state, task.exit_code))

    except Exception as e:
        # Something unexpected happened in the pilot code above
        print("caught Exception: %s" % e)
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        print("need to exit now: %s" % e)

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.
        print("closing session")
        session.close ()

        # the above is equivalent to
        #
        #   session.close (terminate=True)
        #
        # it will thus both clean out the session's database record, and kill
        # all remaining pilots (none in our example).


# -------------------------------------------------------------------------------




__copyright__ = 'Copyright 2016-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import subprocess

import radical.utils as ru

from .base import RMInfo
from .fork import Fork


# ------------------------------------------------------------------------------
#
class Yarn(Fork):

    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, rm_info: RMInfo) -> RMInfo:

        # initialize nodelist from FORK RM
        super()._init_from_scratch(rm_info)

        self.namenode_url = subprocess.check_output(
            ['hdfs', 'getconf', '-nnRpcAddresses']).split(b'\n')[0]
        self._log.debug('Namenode URL = %s', self.namenode_url)

        # I will leave it for the moment because I have not found another way
        # to take the necessary value yet.
        yarn_conf_output = subprocess.check_output(
            ['yarn', 'node', '-list'], stderr=subprocess.STDOUT).split(b'\n')
        for line in yarn_conf_output:
            line = ru.as_string(line)
            if 'ResourceManager' in line:
                settings = line.split('at ')[1]
                if '/' in settings:
                    rm_url = settings.split('/')[1]
                    self.rm_ip = rm_url.split(':')[0]
                    self.rm_port = rm_url.split(':')[1]
                else:
                    self.rm_ip = settings.split(':')[0]
                    self.rm_port = settings.split(':')[1]

        return rm_info


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"

import threading     as mt

import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
class Flux(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, lm_cfg, rm_info, session, prof):

        self._flux_handles = list()
        self._details      = list()

        super().__init__(name, lm_cfg, rm_info, session, prof)

    # --------------------------------------------------------------------------
    #
    def _terminate(self):

        for fh in self._flux_handles:
            fh.reset()


    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, env, env_sh):

        self._log.debug('=== flux init from scratch')
        self._prof.prof('flux_start')

        n_partitions        = self._rm_info.details.n_partitions
        n_nodes             = len(self._rm_info.node_list)
        nodes_per_partition = int(n_nodes / n_partitions)
        threads_per_node    = self._rm_info.cores_per_node  # == hw threads
        gpus_per_node       = self._rm_info.gpus_per_node

        assert n_nodes % n_partitions == 0, \
                'n_nodes %d %% n_partitions %d != 0' % (n_nodes, n_partitions)

        self._log.info('using %d flux partitions [%d nodes]',
                       n_partitions, nodes_per_partition)

        threads = list()

        for n in range(n_partitions):

            thread = mt.Thread(target=self._init_partition,
                               args=[n, nodes_per_partition, threads_per_node,
                                     gpus_per_node])
            thread.start()
            threads.append(thread)

        for thread in threads:
            thread.join()

        self._prof.prof('flux_start_ok')

        lm_info = {'env'          : env,
                   'env_sh'       : env_sh,
                   'n_partitions' : n_partitions,
                   'details'      : self._details}

        return lm_info


    # --------------------------------------------------------------------------
    #
    def _init_partition(self, n, nodes_per_partition, threads_per_node,
                        gpus_per_node):

        fh = ru.FluxHelper()

        self._log.debug('=== starting flux partition %d', n)

        # FIXME: this is a hack for frontier and will only work for slurm
        #        resources.  If Flux is to be used more widely, we need to
        #        pull the launch command from the agent's resource manager.
        launcher = ''
        srun = ru.which('srun')
        if srun:
            launcher = 'srun -n %s -N %d --ntasks-per-node 1 --cpus-per-task=%d --gpus-per-task=%d --export=ALL' \
                       % (nodes_per_partition, nodes_per_partition, threads_per_node, gpus_per_node)

        self._log.debug('=== flux partition %d launcher: %s', n, launcher)

        fh.start_flux(launcher=launcher)

        self._flux_handles.append(fh)
        self._details.append({'flux_uri': fh.uri,
                              'flux_env': fh.env,
                              'flux_uid': fh.uid})

        self._log.debug('=== flux partition %d started', n)


    # --------------------------------------------------------------------------
    #
    def _init_from_info(self, lm_info):

        self._log.debug('=== flux init from info')
        self._prof.prof('flux_reconnect')

        self._env          = lm_info['env']
        self._env_sh       = lm_info['env_sh']
        self._details      = lm_info['details']
        self._n_partitions = lm_info['n_partitions']

        threads = list()
        for details in self._details:

            thread = mt.Thread(target=self._reconnect, args=[details])
            thread.start()
            threads.append(thread)

        for thread in threads:
            thread.join()

        self._prof.prof('flux_reconnect_ok')

    # --------------------------------------------------------------------------
    #
    def _reconnect(self, details):

        fh = ru.FluxHelper(uid=details['flux_uid'])
        fh.connect_flux(uri=details['flux_uri'])
        self._flux_handles.append(fh)


    # --------------------------------------------------------------------------
    #
    def get_partition(self, partition):

        assert partition < self._n_partitions
        return self._flux_handles[partition]


    @property
    def n_partitions(self):
        return self._n_partitions

    def can_launch(self, task):
        raise RuntimeError('method cannot be used on Flux LM')


    def get_launch_cmds(self, task, exec_path):
        raise RuntimeError('method cannot be used on Flux LM')


    def get_launcher_env(self):
        raise RuntimeError('method cannot be used on Flux LM')


    def get_rank_cmd(self):

        return 'export RP_RANK=$FLUX_TASK_RANK\n'


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2013-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import radical.utils as ru

LABEL                  = 'label'
DESCRIPTION            = 'description'
NOTES                  = 'notes'
DEFAULT_SCHEMA         = 'default_schema'
SCHEMAS                = 'schemas'
JOB_MANAGER_ENDPOINT   = 'job_manager_endpoint'
JOB_MANAGER_HOP        = 'job_manager_hop'
FILESYSTEM_ENDPOINT    = 'filesystem_endpoint'
DEFAULT_REMOTE_WORKDIR = 'default_remote_workdir'
DEFAULT_QUEUE          = 'default_queue'
RESOURCE_MANAGER       = 'resource_manager'
AGENT_CONFIG           = 'agent_config'
AGENT_SCHEDULER        = 'agent_scheduler'
AGENT_SPAWNER          = 'agent_spawner'
PRE_BOOTSTRAP_0        = 'pre_bootstrap_0'
PRE_BOOTSTRAP_1        = 'pre_bootstrap_1'
RP_VERSION             = 'rp_version'
VIRTENV                = 'virtenv'
VIRTENV_MODE           = 'virtenv_mode'
PYTHON_DIST            = 'python_dist'
PYTHON_INTERPRETER     = 'python_interpreter'
LAUNCH_METHODS         = 'launch_methods'
LFS_PATH_PER_NODE      = 'lfs_path_per_node'
LFS_SIZE_PER_NODE      = 'lfs_size_per_node'
TASK_TMP               = 'task_tmp'
MEM_PER_NODE           = 'mem_per_node'
CORES_PER_NODE         = 'cores_per_node'
GPUS_PER_NODE          = 'gpus_per_node'
SYSTEM_ARCHITECTURE    = 'system_architecture'
SCATTERED              = 'scattered'

FAKE_RESOURCES         = 'fake_resources'
MANDATORY_ARGS         = 'mandatory_args'
FORWARD_TUNNEL_ENDPOINT = 'forward_tunnel_endpoint'

NEW_SESSION_PER_TASK   = 'new_session_per_task'
TASK_PRE_LAUNCH        = 'task_pre_launch'
TASK_POST_LAUNCH       = 'task_post_launch'
TASK_PRE_EXEC          = 'task_pre_exec'
TASK_POST_EXEC         = 'task_post_exec'

RAPTOR                 = 'raptor'
RAPTOR_HB_DELAY        = 'hb_delay'
RAPTOR_HB_TIMEOUT      = 'hb_timeout'
RAPTOR_HB_FREQUENCY    = 'hb_frequency'

ENDPOINTS_DEFAULT      = {JOB_MANAGER_ENDPOINT: 'fork://localhost/',
                          FILESYSTEM_ENDPOINT : 'file://localhost/'}


# ------------------------------------------------------------------------------
#
class RaptorConfig(ru.TypedDict):

    _schema = {
        RAPTOR_HB_DELAY    : int,
        RAPTOR_HB_TIMEOUT  : int,
        RAPTOR_HB_FREQUENCY: int,
    }

    _defaults = {
        RAPTOR_HB_DELAY    : 5,
        RAPTOR_HB_TIMEOUT  : 500,
        RAPTOR_HB_FREQUENCY: 1000,
    }


# ------------------------------------------------------------------------------
#
class AccessSchema(ru.TypedDict):

    _schema = {
        JOB_MANAGER_ENDPOINT: str,
        JOB_MANAGER_HOP     : str,
        FILESYSTEM_ENDPOINT : str,
    }

    _defaults = {
        JOB_MANAGER_ENDPOINT: None,
        JOB_MANAGER_HOP     : None,
        FILESYSTEM_ENDPOINT : None,
    }


# ------------------------------------------------------------------------------
#
class ResourceConfig(ru.TypedDict):
    '''
    docstrings goes here
    '''

    _schema = {
        LABEL                  : str         ,
        DESCRIPTION            : str         ,
        NOTES                  : str         ,
        DEFAULT_SCHEMA         : str         ,
        SCHEMAS                : {str: AccessSchema},
        RAPTOR                 : RaptorConfig,

        # FIXME: AM - need to resolve since  in Session it is moved into RD
        #        `_get_resource_sandbox` ->  `KeyError: 'filesystem_endpoint'`
        JOB_MANAGER_ENDPOINT   : str         ,
        JOB_MANAGER_HOP        : str         ,
        FILESYSTEM_ENDPOINT    : str         ,

        DEFAULT_REMOTE_WORKDIR : str         ,
        DEFAULT_QUEUE          : str         ,
        RESOURCE_MANAGER       : str         ,
        AGENT_CONFIG           : str         ,
        AGENT_SCHEDULER        : str         ,
        AGENT_SPAWNER          : str         ,
        PRE_BOOTSTRAP_0        : [str]       ,
        PRE_BOOTSTRAP_1        : [str]       ,
        RP_VERSION             : str         ,
        VIRTENV                : str         ,
        VIRTENV_MODE           : str         ,
        PYTHON_DIST            : str         ,
        PYTHON_INTERPRETER     : str         ,
        LAUNCH_METHODS         : {str: None} ,
        LFS_PATH_PER_NODE      : str         ,
        LFS_SIZE_PER_NODE      : int         ,
        TASK_TMP               : str         ,
        MEM_PER_NODE           : int         ,
        CORES_PER_NODE         : int         ,
        GPUS_PER_NODE          : int         ,
        SYSTEM_ARCHITECTURE    : {str: None} ,
        SCATTERED              : bool        ,

        FAKE_RESOURCES         : bool        ,
        MANDATORY_ARGS         : [str]       ,
        FORWARD_TUNNEL_ENDPOINT: str         ,
        NEW_SESSION_PER_TASK   : bool        ,
        TASK_PRE_LAUNCH        : [str]       ,
        TASK_POST_LAUNCH       : [str]       ,
        TASK_PRE_EXEC          : [str]       ,
        TASK_POST_EXEC         : [str]       ,
    }

    _defaults = {
        LABEL                  : ''          ,
        DESCRIPTION            : ''          ,
        NOTES                  : ''          ,
        DEFAULT_SCHEMA         : ''          ,
        SCHEMAS                : list()      ,
        RAPTOR                 : RaptorConfig(),

        # FIXME: AM - need to resolve since in Session it is moved into RD
        #        `_get_resource_sandbox` -> `KeyError: 'filesystem_endpoint'`
        JOB_MANAGER_ENDPOINT   : None        ,
        JOB_MANAGER_HOP        : None        ,
        FILESYSTEM_ENDPOINT    : None        ,

        DEFAULT_REMOTE_WORKDIR : ''          ,
        DEFAULT_QUEUE          : ''          ,
        RESOURCE_MANAGER       : ''          ,
        AGENT_CONFIG           : 'default'   ,
        AGENT_SCHEDULER        : 'CONTINUOUS',
        AGENT_SPAWNER          : 'POPEN'     ,
        PRE_BOOTSTRAP_0        : list()      ,
        PRE_BOOTSTRAP_1        : list()      ,
        RP_VERSION             : 'installed' ,
        VIRTENV                : ''          ,
        VIRTENV_MODE           : 'local'     ,
        PYTHON_DIST            : 'default'   ,
        PYTHON_INTERPRETER     : ''          ,
        LAUNCH_METHODS         : dict()      ,
        LFS_PATH_PER_NODE      : ''          ,
        LFS_SIZE_PER_NODE      : 0           ,
        TASK_TMP               : ''          ,
        MEM_PER_NODE           : 0           ,
        CORES_PER_NODE         : 0           ,
        GPUS_PER_NODE          : 0           ,
        SYSTEM_ARCHITECTURE    : dict()      ,
        SCATTERED              : False       ,

        FAKE_RESOURCES         : False       ,
        MANDATORY_ARGS         : list()      ,
        FORWARD_TUNNEL_ENDPOINT: ''          ,
        NEW_SESSION_PER_TASK   : True        ,
        TASK_PRE_LAUNCH        : list()      ,
        TASK_POST_LAUNCH       : list()      ,
        TASK_PRE_EXEC          : list()      ,
        TASK_POST_EXEC         : list()      ,
    }


# ------------------------------------------------------------------------------



#!/usr/bin/env python3

import radical.pilot as rp


# ------------------------------------------------------------------------------
#
def task_state_cb(task, state):

    print('  task %-30s: %s' % (task.uid, task.state))


# ------------------------------------------------------------------------------
#
@rp.pythontask
def hello_world(msg, sleep):
    import time
    print('hello %s: %.3f' % (msg, time.time()))
    time.sleep(sleep)
    print('hello %s: %.3f' % (msg, time.time()))
    return 'hello %s' % msg


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    session = rp.Session()
    try:
        pd = rp.PilotDescription()
        pd.resource = 'local.localhost'
        pd.runtime  =  10
        pd.cores    = 128

        pmgr = rp.PilotManager(session=session)
        tmgr = rp.TaskManager(session=session)
        tmgr.register_callback(task_state_cb)

        pilot = pmgr.submit_pilots([pd])[0]
        tmgr.add_pilots(pilot)

        raptor = pilot.submit_raptors(rp.TaskDescription(
            {'mode': rp.RAPTOR_MASTER}))[0]

        worker = raptor.submit_workers(rp.TaskDescription(
            {'mode': rp.RAPTOR_WORKER}))[0]

        td_func_1 = rp.TaskDescription({'mode': rp.TASK_FUNCTION,
                                        'function': hello_world('client', 3)})

        td_func_2 = rp.TaskDescription({'mode': rp.TASK_FUNCTION,
                                        'function': 'hello',
                                        'args'    : ['raptor', 3]})
        tasks = raptor.submit_tasks([td_func_1, td_func_2])

        tmgr.wait_tasks([task.uid for task in tasks])

        for task in tasks:
            print('%s [%s]:\n%s' % (task.uid, task.state, task.stdout))

        raptor.rpc('stop')
        tmgr.wait_tasks(raptor.uid)
        print('%s [%s]: %s' % (raptor.uid, raptor.state, raptor.stdout))

    finally:
        session.close(download=False)


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'


import os
import copy
import math
import time
import pprint
import shutil
import tempfile

from collections import defaultdict

import threading          as mt

import radical.utils      as ru

from ... import states    as rps
from ... import constants as rpc
from ... import utils     as rpu

from ...staging_directives import complete_url, expand_staging_directives


# ------------------------------------------------------------------------------
# 'enum' for RP's PMGRraunching types
RP_UL_NAME_SAGA  = "SAGA"
RP_UL_NAME_PSI_J = "PSI_J"


# ------------------------------------------------------------------------------
#
class PilotLauncherBase(object):

    def __init__(self, name, log, prof, state_cb):
        '''
        log:      ru.Logger instance to use
        prof:     ru.Profiler instance to use
        state_cb: rp.pmgr.PNGRLaunchingComponent state callback to invoke on
                  pilot state updates
        '''

        self._name     = name
        self._log      = log
        self._prof     = prof
        self._state_cb = state_cb


    # --------------------------------------------------------------------------
    #
    @property
    def name(self):
        return self._name


    # --------------------------------------------------------------------------
    #
    def can_launch(self, rcfg, pilot):
        '''
        check if the give pilot can be launched on the specified target resource
        '''
        raise NotImplementedError('can_launch not implemented')


    # --------------------------------------------------------------------------
    #
    def launch_pilots(self, rcfg, pilots):
        '''
        rcfg:   resource config for resource to launch pilots topic
        pilots: pilot dictionaries for pilots to launch
                expected to contain `job_dict` as basis for job description
        '''
        raise NotImplementedError('launch_pilots not implemented')


    # --------------------------------------------------------------------------
    #
    def kill_pilots(self, pids):
        '''
        pids:   RP UIDs for pilots to cancel
        '''
        raise NotImplementedError('kill_pilots not implemented')


    # --------------------------------------------------------------------------
    #
    def stop(self):
        '''
        terminate this instance, clean up, destroy any threads etc
        '''

        pass


# ------------------------------------------------------------------------------
#
class PMGRLaunchingComponent(rpu.ClientComponent):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self._uid = ru.generate_id(cfg['owner'] + '.launching.%(counter)s',
                                   ru.ID_CUSTOM)

        super().__init__(cfg, session)
        self._pmgr      = self._owner

        self._pilots    = dict()      # dict for all known pilots
        self._lock      = mt.RLock()  # lock on maipulating the above
        self._sandboxes = dict()      # cache of resource sandbox URLs
        self._cancelled = list()      # list of cancelled pilots

        self._mod_dir   = os.path.dirname(os.path.abspath(__file__))
        self._root_dir  = "%s/../../" % self._mod_dir

        # register input queue
        self.register_input(rps.PMGR_LAUNCHING_PENDING,
                            rpc.PMGR_LAUNCHING_QUEUE, self.work)

        # we don't really have an output queue, as we pass control over the
        # pilot jobs to the resource management system (ResourceManager).
        self._stager = rpu.StagingHelper(self._log)

        self._log.info(ru.get_version([self._mod_dir, self._root_dir]))
        self._rp_version, _, _, _, _ = \
                ru.get_version([self._mod_dir, self._root_dir])


        # load all launcher implementations
        self._launchers = dict()

        from .saga  import PilotLauncherSAGA
        from .psi_j import PilotLauncherPSIJ

        impl = {
            RP_UL_NAME_SAGA : PilotLauncherSAGA,
            RP_UL_NAME_PSI_J: PilotLauncherPSIJ
        }

        exceptions = dict()
        for name in [RP_UL_NAME_PSI_J, RP_UL_NAME_SAGA]:
            try:
                ctor = impl[name]
                self._launchers[name] = ctor(name, self._log, self._prof,
                                             self._state_cb)
            except Exception as e:
                self._log.warn('skip launcher %s' % name)
                exceptions[name] = e

        # if no launcher is usable, log the found exceptions
        if not self._launchers:
            for name in [RP_UL_NAME_PSI_J, RP_UL_NAME_SAGA]:
                e = exceptions.get(name)
                if e:
                    try   : raise e
                    except: self._log.exception('launcher %s unusable' % name)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def create(cls, cfg, session):

        return cls(cfg, session)


    # --------------------------------------------------------------------------
    #
    def _state_cb(self, pilot, rp_state):

        self._log.info('pilot state update: %s: %s', pilot['uid'], rp_state)

        if rp_state != pilot['state']:
            self.advance(pilot, state=rp_state, push=False, publish=True)


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        # FIXME: is this called?

        try:
            self.unregister_input(rps.PMGR_LAUNCHING_PENDING,
                                  rpc.PMGR_LAUNCHING_QUEUE, self.work)

            # FIXME: always kill all pilot jobs for non-final pilots at
            #        termination, and set the pilot states to CANCELED.
            with self._lock:
                pids = list(self._pilots.keys())

            self._kill_pilots(pids)

            # TODO: close launchers

        except:
            self._log.exception('finalization error')


    # --------------------------------------------------------------------------
    #
    def control_cb(self, topic, msg):

        cmd = msg['cmd']
        arg = msg['arg']

        self._log.debug_9('launcher got %s', msg)

        if cmd == 'kill_pilots':

            pmgr = arg['pmgr']
            pids = arg['uids']

            if pmgr != self._pmgr:
                return True

            if not isinstance(pids, list):
                pids = [pids]

            self._log.info('received "kill_pilots" command (%s)', pids)

            self._kill_pilots(pids)


    # --------------------------------------------------------------------------
    #
    def _kill_pilots(self, pids):
        '''
        Send a cancellation request to the pilots.  This call will not wait for
        the request to get enacted, nor for it to arrive, but just send it.
        '''

        if not pids:
            if not self._pilots:
                return
            else:
                pids = list(self._pilots.keys())

        with self._lock:

            for pid in pids:

                self._log.debug('cancel pilot %s', pid)
                if pid not in self._pilots:
                    self._log.warn('cannot cancel unknown pilot %s', pid)
                    self._cancelled.append(pid)
                    continue

                pilot = self._pilots[pid]
                lname = pilot['launcher']

                if lname not in self._launchers:
                    self._log.warn('invalid pilot launcher name: %s', lname)
                    continue

                launcher = self._launchers[lname]
                try:
                    launcher.kill_pilots([pid])
                except:
                    self._log.exception('pilot cancel failed for %s' % pid)


    # --------------------------------------------------------------------------
    #
    def work(self, pilots):

        pilots = ru.as_list(pilots)

        # weed out pilots for which we have already received a cancel request
        to_cancel = [p for p in pilots if p['uid']     in self._cancelled]
        to_start  = [p for p in pilots if p['uid'] not in self._cancelled]

        self.advance(to_cancel, rps.CANCELED,       publish=True, push=False)
        self.advance(to_start,  rps.PMGR_LAUNCHING, publish=True, push=False)

        # We can only use bulk submission for pilots which go to the same
        # target, thus we sort them into buckets and launch the buckets
        # individually
        buckets = defaultdict(lambda: defaultdict(list))
        for pilot in to_start:
            resource = pilot['description']['resource']
            schema   = pilot['description']['access_schema']
            buckets[resource][schema].append(pilot)

        for resource in buckets:

            for schema in buckets[resource]:

                try:
                    pilots = buckets[resource][schema]
                    pids   = [p['uid'] for p in pilots]
                    self._log.info("Launching pilots on %s: %s", resource, pids)

                    self._start_pilot_bulk(resource, schema, pilots)

                    self.advance(pilots, rps.PMGR_ACTIVE_PENDING,
                                         push=False, publish=True)

                except Exception:
                    self._log.exception('bulk launch failed')
                    self.advance(pilots, rps.FAILED, push=False, publish=True)


    # --------------------------------------------------------------------------
    #
    def _start_pilot_bulk(self, resource, schema, pilots):
        '''
        For each pilot, we prepare by determining what files need to be staged,
        and what job description needs to be submitted.  Files are then be
        staged, and jobs are launched.

        Two files are staged: a bootstrapper and a tarball - the latter
        containing the pilot sandboxes, agent configs, and any other auxilliary
        files needed to bootstrap.  The bootstrapper will untar those parts of
        the tarball which it needs to bootstrap one specific pilot.
        '''

        rcfg = self._session.get_resource_config(resource, schema)
        sid  = self._session.uid

        # ----------------------------------------------------------------------
        # the rcfg can contain keys with string expansion placeholders where
        # values from the pilot description need filling in.  A prominent
        # example is `%(pd.project)s`, where the pilot description's `PROJECT`
        # value needs to be filled in (here in lowercase).
        #
        # FIXME: right now we assume all pilot descriptions to contain similar
        #        entries, so that the expansion is only done on the first PD.
        expand = dict()
        pd     = pilots[0]['description']
        for k,v in pd.items():
            if v is None:
                v = ''
            expand['pd.%s' % k] = v
            if isinstance(v, str):
                expand['pd.%s' % k.upper()] = v.upper()
                expand['pd.%s' % k.lower()] = v.lower()
            else:
                expand['pd.%s' % k.upper()] = v
                expand['pd.%s' % k.lower()] = v

        for k in rcfg:
            if isinstance(rcfg[k], str):
                orig     = rcfg[k]
                rcfg[k]  = rcfg[k] % expand
                expanded = rcfg[k]
                if orig != expanded:
                    self._log.debug('RCFG:\n%s\n%s', orig, expanded)

        # we create a fake session_sandbox with all pilot_sandboxes in /tmp, and
        # then tar it up.  Once we untar that tarball on the target machine, we
        # should have all sandboxes and all files required to bootstrap the
        # pilots
        tmp_dir  = os.path.abspath(tempfile.mkdtemp(prefix='rp_agent_tmp'))
        tar_name = '%s.%s.tgz' % (sid, self._uid)
        tar_tgt  = '%s/%s'     % (tmp_dir, tar_name)
        tar_url  = ru.Url('file://localhost/%s' % tar_tgt)

        # we need the session sandbox url, but that is (at least in principle)
        # dependent on the schema to use for pilot startup.  So we confirm here
        # that the bulk is consistent wrt. to the schema.  Also include
        # `staging_input` files and place them in the `pilot_sandbox`.
        #
        # FIXME: if it is not, it needs to be splitted into schema-specific
        # sub-bulks
        #
        schema = pd.get('access_schema')
        for pilot in pilots[1:]:
            assert schema == pilot['description'].get('access_schema'), \
                    'inconsistent scheme on launch / staging'

        # get and expand sandboxes (this bulk uses the same schema toward the
        # same target resource, so all session sandboxes are the same)
        # FIXME: expansion actually may differ per pilot (queue names, project
        #        names, etc could be expanded)
        session_sandbox = self._session._get_session_sandbox(pilots[0]).path
        session_sandbox = session_sandbox % expand

        # we will create the session sandbox before we untar, so we can use that
        # as workdir, and pack all paths relative to that session sandbox.  That
        # implies that we have to recheck that all URLs in fact do point into
        # the session sandbox.
        #
        # We also create a file `staging_output.json` for each pilot which
        # contains the list of files to be tarred up and prepared for output
        # staging.

        ft_list = list()  # files to stage
        sd_list = list()  # staging directives

        for pilot in pilots:

            pid = pilot['uid']
            os.makedirs('%s/%s' % (tmp_dir, pid))

            self._prepare_pilot(resource, rcfg, pilot, expand, tar_name)

            ft_list += pilot['fts']
            sd_list += pilot['sds']


        # ----------------------------------------------------------------------
        # handle pilot data staging
        for pilot in pilots:

            pid = pilot['uid']
            self._prof.prof('staging_in_start', uid=pid)

            for sd in ru.as_list(pilot['description'].get('input_staging')):

                if isinstance(sd, dict):
                    sd['rem'] = False
                    ft_list.append(sd)

                else:
                    fname = sd
                    base = os.path.basename(fname)
                    # checking if input staging file exists
                    if fname.startswith('./'):
                        fname = fname.split('./', maxsplit=1)[1]
                    if not fname.startswith('/'):
                        fname = os.path.join(self._cfg.base, fname)
                    if not os.path.exists(fname):
                        raise RuntimeError('input_staging file does not exists: '
                                           '%s for pilot %s' % (fname, pid))

                    ft_list.append({'source': fname,
                                    'target': '%s/%s' % (pid, base),
                                    'rem': False})

            output_staging = pilot['description'].get('output_staging')
            if output_staging:
                fname = '%s/%s/staging_output.txt' % (tmp_dir, pilot['uid'])
                with ru.ru_open(fname, 'w') as fout:
                    for entry in output_staging:
                        fout.write('%s\n' % entry)

        # direct staging, use first pilot for staging context
        # NOTE: this implies that the SDS can only refer to session
        #       sandboxes, not to pilot sandboxes!
        self._stage_in(pilots[0], sd_list)


        for ft in ft_list:
            source     = os.path.abspath(ft['source'])
            target     = os.path.relpath(os.path.normpath(ft['target']), session_sandbox)
          # source_dir = os.path.dirname(source)
            target_dir = os.path.dirname(target)

            if target_dir.startswith('..'):
                target = ft['target']
                target_dir = os.path.dirname(target)

            if not os.path.isdir('%s/%s' % (tmp_dir, target_dir)):
                os.makedirs('%s/%s' % (tmp_dir, target_dir))

            if source == '/dev/null':
                # we want an empty file -- touch it (tar will refuse to
                # handle a symlink to /dev/null)
                ru.ru_open('%s/%s' % (tmp_dir, target), 'a').close()
            else:
                # use a shell callout to account for wildcard expansion
                cmd = 'ln -s %s %s/%s' % (os.path.abspath(source), tmp_dir,
                                          target)
                out, err, ret = ru.sh_callout(cmd, shell=True)
                if ret:
                    self._log.debug('cmd: %s', cmd)
                    self._log.debug('out: %s', out)
                    self._log.debug('err: %s', err)
                    raise RuntimeError('callout failed: %s' % cmd)


        # tar.  If any command fails, this will raise.
        cmd = "cd %s && tar zchf %s *" % (tmp_dir, tar_tgt)
        out, err, ret = ru.sh_callout(cmd, shell=True)

        if ret:
            self._log.debug('cmd: %s', cmd)
            self._log.debug('out: %s', out)
            self._log.debug('err: %s', err)
            raise RuntimeError('callout failed: %s' % cmd)

        # remove all files marked for removal-after-pack
        for ft in ft_list:
            if ft['rem']:
                os.unlink(ft['source'])

        fs_endpoint  = rcfg['filesystem_endpoint']
        fs_url       = ru.Url(fs_endpoint)
        tar_rem      = ru.Url(fs_url)
        tar_rem.path = "%s/%s" % (session_sandbox, tar_name)

        self._log.debug('stage tarball for %s', pilots[0]['uid'])
        self._stage_in(pilots[0], {'source': tar_url,
                                   'target': tar_rem,
                                   'action': rpc.TRANSFER})
        shutil.rmtree(tmp_dir)

        # FIXME: the untar was moved into the bootstrapper (see `-z`).  That
        #        is actually only correct for the single-pilot case...

        now = time.time()
        for pilot in pilots:
            self._prof.prof('staging_in_stop',  uid=pilot['uid'], ts=now)
            self._prof.prof('submission_start', uid=pilot['uid'], ts=now)

        # find launchers to handle pilot job submission.  We sort by launcher so
        # that each launcher can handle a bulk of pilots at once.
        buckets = defaultdict(list)
        for pilot in pilots:
            for lname,launcher in self._launchers.items():
                if launcher.can_launch(rcfg, pilots):
                    pilot['launcher'] = lname
                    buckets[lname].append(pilot)
                    self._log.info('use launcher %s for pilot %s',
                                   lname, pilot['uid'])
                    break

            if not pilot.get('launcher'):
                raise RuntimeError('no launcher found for %s' % pilot['uid'])

        with self._lock:
            for lname, bucket in buckets.items():
                launcher = self._launchers[lname]
                launcher.launch_pilots(rcfg, bucket)
                for pilot in bucket:
                    pid = pilot['uid']
                    self._pilots[pid] = pilot
                    self._prof.prof('submission_stop', uid=pid)


    # --------------------------------------------------------------------------
    #
    def _prepare_pilot(self, resource, rcfg, pilot, expand, tar_name):

        rcfg.verify()

        pid = pilot["uid"]
        pilot['fts'] = list()  # tar for staging
        pilot['sds'] = list()  # direct staging
        pilot['jd_dict' ] = None    # job description

        # ----------------------------------------------------------------------
        # Database connection parameters
        sid       = self._session.uid
        proxy_url = self._session.cfg.proxy_url

        # ----------------------------------------------------------------------
        # pilot description and resource configuration
        requested_nodes  = pilot['description']['nodes']
        requested_cores  = pilot['description']['cores']
        requested_gpus   = pilot['description']['gpus']
        requested_memory = pilot['description']['memory']
        runtime          = pilot['description']['runtime']
        app_comm         = pilot['description']['app_comm']
        queue            = pilot['description']['queue']
        job_name         = pilot['description']['job_name']
        project          = pilot['description']['project']
        cleanup          = pilot['description']['cleanup']
        services         = pilot['description']['services']
        enable_ep        = pilot['description']['enable_ep']
        prepare_env      = pilot['description']['prepare_env']
        reconfig_src     = pilot['description']['reconfig_src']

        # ----------------------------------------------------------------------
        # get parameters from resource cfg, set defaults where needed
        agent_spawner           = rcfg.agent_spawner
        agent_config            = rcfg.agent_config
        agent_scheduler         = rcfg.agent_scheduler
        default_queue           = rcfg.default_queue
        forward_tunnel_endpoint = rcfg.forward_tunnel_endpoint
        resource_manager        = rcfg.resource_manager
        pre_bootstrap_0         = rcfg.pre_bootstrap_0
        pre_bootstrap_1         = rcfg.pre_bootstrap_1
        python_interpreter      = rcfg.python_interpreter
        rp_version              = rcfg.rp_version
        virtenv_mode            = rcfg.virtenv_mode
        virtenv                 = rcfg.virtenv
        cores_per_node          = rcfg.cores_per_node
        gpus_per_node           = rcfg.gpus_per_node
        lfs_path_per_node       = rcfg.lfs_path_per_node
        lfs_size_per_node       = rcfg.lfs_size_per_node
        python_dist             = rcfg.python_dist
        task_tmp                = rcfg.task_tmp
        task_pre_launch         = rcfg.task_pre_launch
        task_post_launch        = rcfg.task_post_launch
        task_pre_exec           = rcfg.task_pre_exec
        task_post_exec          = rcfg.task_post_exec
        mandatory_args          = rcfg.mandatory_args
        system_architecture     = rcfg.system_architecture
        raptor_cfg              = rcfg.raptor

        # part of the core specialization settings
        blocked_cores           = system_architecture.get('blocked_cores', [])
        blocked_gpus            = system_architecture.get('blocked_gpus',  [])

        # some default values are determined at runtime
        if not virtenv:
            virtenv = '%%(resource_sandbox)s/ve.%s.%s' \
                    % (resource, self._rp_version)

        self._log.debug(pprint.pformat(rcfg))

        # make sure that mandatory args are known
        for ma in mandatory_args:
            if pilot['description'].get(ma) is None:
                raise  ValueError('attribute "%s" is required for "%s"'
                                 % (ma, resource))

        # get pilot and global sandbox
        endpoint_fs      = self._session._get_endpoint_fs     (pilot)
        resource_sandbox = self._session._get_resource_sandbox(pilot)
        session_sandbox  = self._session._get_session_sandbox (pilot)
        pilot_sandbox    = self._session._get_pilot_sandbox   (pilot)
        client_sandbox   = self._session._get_client_sandbox  ()

        pilot['endpoint_fs']      = str(endpoint_fs)      % expand
        pilot['resource_sandbox'] = str(resource_sandbox) % expand
        pilot['session_sandbox']  = str(session_sandbox)  % expand
        pilot['pilot_sandbox']    = str(pilot_sandbox)    % expand
        pilot['client_sandbox']   = str(client_sandbox)

        # from here on we need only paths
        endpoint_fs      = endpoint_fs     .path % expand
        resource_sandbox = resource_sandbox.path % expand
        session_sandbox  = session_sandbox .path % expand
        pilot_sandbox    = pilot_sandbox   .path % expand
      # client_sandbox   = client_sandbox  # not expanded

        # expand variables in virtenv string
        virtenv = virtenv % {'pilot_sandbox'   : pilot_sandbox,
                             'session_sandbox' : session_sandbox,
                             'resource_sandbox': resource_sandbox}

        if not job_name:
            job_name = pid

        try:
            if isinstance(agent_config, dict):
                agent_cfg = ru.Config(cfg=agent_config)

            elif isinstance(agent_config, str):
                agent_cfg = ru.Config('radical.pilot',
                                      category='agent',
                                      name=agent_config)
            else:
                # we can't handle this type
                raise TypeError('agent config must be string or dict')

        except Exception:
            self._log.exception('Error using agent config')
            raise


        # Check for deprecated global_virtenv
        if 'global_virtenv' in rcfg:
            raise RuntimeError("'global_virtenv' is deprecated (%s)" % resource)

        # Create a host:port string for use by the bootstrap_0.
        tmp = ru.Url(proxy_url)
        if tmp.port:
            hostport = "%s:%d" % (tmp.host, tmp.port)
        else:
            hostport = tmp.host

        # ----------------------------------------------------------------------
        # the version of the agent is derived from
        # rp_version, which has the following format
        # and interpretation:
        #
        # case rp_version:
        #   @<token>:
        #   @tag/@branch/@commit:
        #       git clone $github_base radical.pilot.src
        #       (cd radical.pilot.src && git checkout token)
        #       pip install -t $VIRTENV/rp_install/ radical.pilot.src
        #       rm -rf radical.pilot.src
        #       export PYTHONPATH=$VIRTENV/rp_install:$PYTHONPATH
        #
        #   release:
        #       pip install -t $VIRTENV/rp_install radical.pilot
        #       export PYTHONPATH=$VIRTENV/rp_install:$PYTHONPATH
        #
        #   installed:
        #       true
        # esac
        #
        # virtenv_mode
        #   private : error  if ve exists, otherwise create, then use
        #   update  : update if ve exists, otherwise create, then use
        #   create  : use    if ve exists, otherwise create, then use
        #   use     : use    if ve exists, otherwise error,  then exit
        #   recreate: delete if ve exists, otherwise create, then use
        #   local   : use the client virtualenv (assumes same FS)
        #
        # examples   :
        #   virtenv@v0.20
        #   virtenv@devel
        #   virtenv@release
        #   virtenv@installed
        #   stage@local
        #   stage@/tmp/my_agent.py
        #
        # Note that some combinations may be invalid,
        # specifically in the context of virtenv_mode.  If, for
        # example, virtenv_mode is 'use', then the 'virtenv:tag'
        # will not make sense, as the virtenv is not updated.
        # In those cases, the virtenv_mode is honored, and
        # a warning is printed.
        #
        # Also, the 'stage' mode can only be combined with the
        # 'local' source, or with a path to the agent (relative
        # to root_dir, or absolute).
        #
        # A rp_version which does not adhere to the
        # above syntax is ignored, and the fallback stage@local
        # is used.

        # use local VE ?
        if virtenv_mode == 'local':
            rp_version = 'installed'
            if os.environ.get('VIRTUAL_ENV'):
                python_dist = 'default'
                virtenv     = os.environ['VIRTUAL_ENV']
            elif os.environ.get('CONDA_PREFIX'):
                python_dist = 'anaconda'
                virtenv     = os.environ['CONDA_PREFIX']
            else:
                # we can't use local
                self._log.error('virtenv_mode is local, no local env found')
                raise ValueError('no local env found')

        if not rp_version.startswith('@') and \
               rp_version not in ['installed', 'local', 'release']:
            raise ValueError("invalid rp_version '%s'" % rp_version)

        if rp_version.startswith('@'):
            rp_version  = rp_version[1:]  # strip '@'

        # ----------------------------------------------------------------------
        # sanity checks
        RE = RuntimeError
        if not python_dist     : raise RE("missing python distribution")
        if not agent_spawner   : raise RE("missing agent spawner")
        if not agent_scheduler : raise RE("missing agent scheduler")
        if not resource_manager: raise RE("missing resource manager")

        # massage some values
        if not queue:
            queue = default_queue

        if  cleanup and isinstance(cleanup, bool):
            #  l : log files
            #  u : task work dirs
            #  v : virtualenv
            #  e : everything (== pilot sandbox)
            cleanup = 'luve'

            # we never cleanup virtenvs which are not private
            if virtenv_mode != 'private':
                cleanup = cleanup.replace('v', '')

        # estimate requested resources
        smt = int(os.environ.get('RADICAL_SMT') or
                  system_architecture.get('smt', 1))

        if cores_per_node and smt:
            cores_per_node *= smt

        avail_cores_per_node = cores_per_node
        avail_gpus_per_node  = gpus_per_node

        if avail_cores_per_node and blocked_cores:
            avail_cores_per_node -= len(blocked_cores)
            assert (avail_cores_per_node > 0)

        if avail_gpus_per_node and blocked_gpus:
            avail_gpus_per_node -= len(blocked_gpus)
            assert (avail_gpus_per_node >= 0)

        if not requested_nodes and not requested_cores:
            requested_nodes = 1

        if requested_nodes:

            if not avail_cores_per_node:
                raise RuntimeError('use "cores" in PilotDescription')

        else:

            if avail_cores_per_node:
                requested_nodes = requested_cores / avail_cores_per_node

            if avail_gpus_per_node:
                requested_nodes = max(requested_gpus / avail_gpus_per_node,
                                      requested_nodes)

            requested_nodes = math.ceil(requested_nodes)

        # now that we know the number of nodes to request, derive
        # the *actual* number of cores and gpus we allocate
        allocated_cores = (
            requested_nodes * avail_cores_per_node) or requested_cores
        allocated_gpus  = (
            requested_nodes * avail_gpus_per_node)  or requested_gpus

        self._log.debug('nodes: %s [%s %s], cores: %s, gpus: %s',
                        requested_nodes, cores_per_node, gpus_per_node,
                        allocated_cores, allocated_gpus)

        # set mandatory args
        bs_args = ['-l', '%s/bootstrap_0.sh' % pilot_sandbox]

        bs_args.extend(['-p', pid])
        bs_args.extend(['-s', sid])
        bs_args.extend(['-m', virtenv_mode])
        bs_args.extend(['-r', rp_version])
        bs_args.extend(['-b', python_dist])
        bs_args.extend(['-v', virtenv])
        bs_args.extend(['-y', str(runtime)])
        bs_args.extend(['-z', tar_name])

        # set optional args
        if resource_manager == "CCM": bs_args.extend(['-c'])
        if forward_tunnel_endpoint:   bs_args.extend(['-f', forward_tunnel_endpoint])
        if forward_tunnel_endpoint:   bs_args.extend(['-h', hostport])
        if python_interpreter:        bs_args.extend(['-i', python_interpreter])
      # if tunnel_bind_device:        bs_args.extend(['-t', tunnel_bind_device])
        if cleanup:                   bs_args.extend(['-x', cleanup])

        for arg in pre_bootstrap_0:   bs_args.extend(['-e', arg])
        for arg in pre_bootstrap_1:   bs_args.extend(['-w', arg])

        agent_cfg['uid']                 = 'agent_0'
        agent_cfg['sid']                 = sid
        agent_cfg['pid']                 = pid
        agent_cfg['owner']               = pid
        agent_cfg['pmgr']                = self._pmgr
        agent_cfg['resource']            = resource
        agent_cfg['nodes']               = requested_nodes
        agent_cfg['cores']               = allocated_cores
        agent_cfg['gpus']                = allocated_gpus
        agent_cfg['spawner']             = agent_spawner
        agent_cfg['scheduler']           = agent_scheduler
        agent_cfg['runtime']             = runtime
        agent_cfg['app_comm']            = app_comm
        agent_cfg['proxy_url']           = proxy_url
        agent_cfg['pilot_sandbox']       = pilot_sandbox
        agent_cfg['session_sandbox']     = session_sandbox
        agent_cfg['resource_sandbox']    = resource_sandbox
        agent_cfg['resource_manager']    = resource_manager
        agent_cfg['cores_per_node']      = cores_per_node
        agent_cfg['gpus_per_node']       = gpus_per_node
        agent_cfg['lfs_path_per_node']   = lfs_path_per_node
        agent_cfg['lfs_size_per_node']   = lfs_size_per_node
        agent_cfg['task_tmp']            = task_tmp
        agent_cfg['task_pre_launch']     = task_pre_launch
        agent_cfg['task_pre_exec']       = task_pre_exec
        agent_cfg['task_post_launch']    = task_post_launch
        agent_cfg['task_post_exec']      = task_post_exec
        agent_cfg['resource_cfg']        = copy.deepcopy(rcfg)
        agent_cfg['log_lvl']             = self._log.level
        agent_cfg['debug_lvl']           = self._log.debug_level
        agent_cfg['services']            = services
        agent_cfg['enable_ep']           = enable_ep
        agent_cfg['prepare_env']         = prepare_env
        agent_cfg['reconfig_src']        = reconfig_src
        agent_cfg['raptor']              = raptor_cfg

        pilot['cfg']       = agent_cfg
        pilot['resources'] = {'cpu': allocated_cores,
                              'gpu': allocated_gpus}


        # ----------------------------------------------------------------------
        # Write agent config dict to a json file in pilot sandbox.

        agent_cfg_name = 'agent_0.cfg'
        cfg_tmp_handle, cfg_tmp_file = tempfile.mkstemp(prefix='rp.agent_cfg.')
        os.close(cfg_tmp_handle)  # file exists now

        # Convert dict to json file
        self._log.debug("Write agent cfg to '%s'.", cfg_tmp_file)
        agent_cfg.write(cfg_tmp_file)

        # always stage agent cfg for each pilot, not in the tarball
        # FIXME: purge the tmp file after staging
        self._log.debug('cfg %s -> %s', agent_cfg['pid'], pilot_sandbox)
        pilot['sds'].append({'source': cfg_tmp_file,
                             'target': '%s/%s' % (pilot['pilot_sandbox'],
                                                  agent_cfg_name),
                             'action': rpc.TRANSFER})

        # always stage the bootstrapper for each pilot, not in the tarball
        # FIXME: this results in many staging ops for many pilots
        bootstrapper_path = os.path.abspath("%s/agent/bootstrap_0.sh"
                                           % self._root_dir)
        pilot['sds'].append({'source': bootstrapper_path,
                             'target': '%s/bootstrap_0.sh'
                                     % pilot['pilot_sandbox'],
                             'action': rpc.TRANSFER})

        # always stage RU env helper
        env_helper = ru.which('radical-utils-env.sh')
        assert env_helper
        self._log.debug('env %s -> %s', env_helper, pilot_sandbox)
        pilot['sds'].append({'source': env_helper,
                             'target': '%s/%s' % (pilot['pilot_sandbox'],
                                                  os.path.basename(env_helper)),
                             'action': rpc.TRANSFER})

        # check if we have a sandbox cached for that resource.  If so, we have
        # nothing to do.  Otherwise we create the sandbox and stage the RP
        # stack etc.
        #
        # NOTE: this will race when multiple pilot launcher instances are used!
        #
        if resource not in self._sandboxes:

            self._sandboxes[resource] = True

        # ----------------------------------------------------------------------
        # Create Job description

        jd_dict = ru.TypedDict()

        jd_dict.name                  = job_name
        jd_dict.executable            = '/bin/bash'
        jd_dict.arguments             = bs_args
        jd_dict.working_directory     = pilot_sandbox
        jd_dict.project               = project
        jd_dict.output                = 'bootstrap_0.out'
        jd_dict.error                 = 'bootstrap_0.err'
        jd_dict.node_count            = requested_nodes
        jd_dict.total_cpu_count       = allocated_cores
        jd_dict.total_gpu_count       = allocated_gpus
        jd_dict.total_physical_memory = requested_memory
        jd_dict.processes_per_host    = avail_cores_per_node
        jd_dict.wall_time_limit       = runtime
        jd_dict.queue                 = queue
        jd_dict.file_transfer         = list()
        jd_dict.environment           = dict()
        jd_dict.system_architecture   = dict(system_architecture)

        # job description environment variable(s) setup
        if self._prof.enabled:
            jd_dict.environment['RADICAL_PROFILE'] = 'TRUE'

        jd_dict.environment['RP_PILOT_SANDBOX'] = pilot_sandbox
        jd_dict.environment['RADICAL_BASE']     = resource_sandbox
        jd_dict.environment['RADICAL_SMT']      = str(smt)

        # for condor backends and the like which do not have shared FSs, we add
        # additional staging directives so that the backend system binds the
        # files from the session and pilot sandboxes to the pilot job.

        self._log.debug("Bootstrap command line: %s %s", jd_dict.executable,
                jd_dict.arguments)

        pilot['jd_dict'] = jd_dict


    # --------------------------------------------------------------------------
    #
    def _stage_in(self, pilot, sds):
        '''
        Run some input staging directives.
        '''

        # contexts for staging url expansion
        tgt_ctx = {'pwd'     : pilot['pilot_sandbox'],      # !
                   'client'  : pilot['client_sandbox'],
                   'pilot'   : pilot['pilot_sandbox'],
                   'resource': pilot['resource_sandbox']}

        src_ctx = {'pwd'     : pilot['client_sandbox'],     # !
                   'client'  : pilot['client_sandbox'],
                   'pilot'   : pilot['pilot_sandbox'],
                   'resource': pilot['resource_sandbox']}

        sds = expand_staging_directives(sds, src_ctx, tgt_ctx)

        for sd in sds:
            self._prof.prof('staging_in_start', uid=pilot['uid'], msg=sd['uid'])
            self._stager.handle_staging_directive(sd)
            self._prof.prof('staging_in_stop', uid=pilot['uid'], msg=sd['uid'])


    # --------------------------------------------------------------------------
    #
    def _stage_out(self, pilot, sds):
        '''
        Run some output staging directives.
        '''

        # contexts for staging url expansion
        loc_ctx = {'pwd'     : pilot['client_sandbox'],
                   'client'  : pilot['client_sandbox'],
                   'pilot'   : pilot['pilot_sandbox'],
                   'resource': pilot['resource_sandbox']}

        rem_ctx = {'pwd'     : pilot['pilot_sandbox'],
                   'client'  : pilot['client_sandbox'],
                   'pilot'   : pilot['pilot_sandbox'],
                   'resource': pilot['resource_sandbox']}

        sds = ru.as_list(sds)

        for sd in sds:
            sd['source'] = str(complete_url(sd['source'], rem_ctx, self._log))
            sd['target'] = str(complete_url(sd['target'], loc_ctx, self._log))

            self._prof.prof('staging_out_start', uid=pilot['uid'], msg=sd['uid'])
            self._stager.handle_staging_directive(sd)
            self._prof.prof('staging_out_stop', uid=pilot['uid'], msg=sd['uid'])


# ------------------------------------------------------------------------------





'''
The serializer should be able to (de)serialize information that we want
to send over the wire from the client side to the agent side via
1- ZMQ
2- MongoDB

we except:
    1- Callables with and without dependecies.
    2- Non-callables like classes and other python objects
'''
__copyright__ = 'Copyright 2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os
import dill
import pickle
import codecs
import tempfile


_obj_dir       = tempfile.gettempdir()
_obj_file_name = 'rp_obj.pkl'
_obj_file_path = os.path.join(_obj_dir, _obj_file_name)


# ------------------------------------------------------------------------------
#
def serialize_obj(obj):
    '''
    serialize object
    '''

    if callable(obj):
        try:
            return dill.dumps(obj)
        except:
            # if we fail, then pickle it by reference
            # see issue: https://github.com/uqfoundation/dill/issues/128
            return dill.dumps(obj, byref = True)

    else:
        try:
            return dill.dumps(obj, recurse = True)
        except:
            # if we fail, then pickle it by reference
            # see issue: https://github.com/uqfoundation/dill/issues/128
            return dill.dumps(obj, byref = True)


# ------------------------------------------------------------------------------
#
def serialize_file(obj, fname=None):
    '''
    serialize object to file
    '''

    if not fname:
        fname = _obj_file_path

    # FIXME: assign unique path and id for the pickled file
    #        to avoid overwriting to the same file
    with open(fname, 'wb') as f:
        f.write(serialize_obj(obj))

    return _obj_file_path


# ------------------------------------------------------------------------------
#
def deserialize_file(fname):
    '''
    Deserialize object from file
    '''

    with open(fname, 'rb') as f:
        return dill.load(f)


# ------------------------------------------------------------------------------
#
def deserialize_obj(data):
    '''
    Deserialize object from str.
    '''

    return dill.loads(data)


# ------------------------------------------------------------------------------
#
def serialize_bson(obj):

    return codecs.encode(pickle.dumps(obj), "base64").decode()


# ------------------------------------------------------------------------------
#
def deserialize_bson(obj):

    return pickle.loads(codecs.decode(obj.encode(), "base64"))


# ------------------------------------------------------------------------------




# pylint: disable=import-error

import io
import os
import sys
import time

import threading           as mt
import radical.utils       as ru

from .worker            import Worker

from ..states           import AGENT_EXECUTING_PENDING, AGENT_EXECUTING


# MPI message tags
TAG_REGISTER_REQUESTS    = 110
TAG_REGISTER_REQUESTS_OK = 111
TAG_SEND_TASK            = 120
TAG_RECV_RESULT          = 121
TAG_REGISTER_RESULTS     = 130
TAG_REGISTER_RESULTS_OK  = 131

# message payload constants
MSG_PING = 210
MSG_PONG = 211
MSG_OK   = 220
MSG_NOK  = 221

# resource allocation flags
FREE = 0
BUSY = 1


# ------------------------------------------------------------------------------
#
class _Resources(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, log, prof, ranks):

        self._log   = log
        self._prof  = prof
        self._ranks = ranks

        # FIXME: RP considers MPI tasks to be homogeneous, in that all ranks of
        #        the task have the same set of resources allocated.  That
        #        implies that all ranks in this MPI worker have the same
        #        resources allocated.  That implies that, in most cases, no rank
        #        has GPUs alloated (we place one rank per core, but the number
        #        of cores per node is in general different than the number of
        #        GPUs per node).
        #
        #        RP will need to support heterogeneous MPI tasks to allow this
        #        worker to also assign GPUs to specific ranks.
        #
        self._res_evt   = mt.Event()  # signals free resources
        self._res_lock  = mt.Lock()   # lock resource for alloc / dealloc
        self._resources = {
                'cores': [0] * self._ranks,
              # 'gpus' : [0] * self._n_gpus
        }

        # resources are initially all free
        self._res_evt.set()


    # --------------------------------------------------------------------------
    #
    def __str__(self):

        out = ':'
        for r in self._resources['cores']:
            if r == FREE: out += '-'
            else        : out += '#'
        out += ':'
      # for r in self._resources['gpus']:
      #     if r == FREE: out += '-'
      #     else        : out += '#'
      # out += ':'
        return out


    # --------------------------------------------------------------------------
    #
    @property
    def log(self): return self._log

    @property
    def prof(self): return self._prof

    @property
    def ranks(self): return self._ranks


    # --------------------------------------------------------------------------
    #
    def _alloc(self, task):
        '''
        This call will search for free cores and gpus to run the task.  More
        precisely, the core will wait for a sufficient number of ranks to become
        available whose resources are suitable to run the task.  The call will
        block until those ranks are found.
        '''

        # FIXME: handle threads
        # FIXME: handle GPUs

        uid = task['uid']

        self._log.debug_5('alloc %s', uid)
        self._prof.prof('schedule_try', uid=uid)

        cores = task['description'].get('ranks', 1)

        if cores > self._ranks:
            raise ValueError('insufficient resources to run task (%d > %d'
                    % (cores, self._ranks))

        self._log.debug_5('alloc %s: %s', task['uid'], cores)

        while True:

            if self._res_evt.is_set():

                with self._res_lock:

                    if cores > self._resources['cores'].count(FREE):
                        self._res_evt.clear()
                        continue

                    ranks = list()
                    for rank in range(self._ranks):

                        if self._resources['cores'][rank] == FREE:

                            self._resources['cores'][rank] = BUSY
                            ranks.append(rank)

                            if len(ranks) == cores:
                                self._prof.prof('schedule_ok', uid=uid)
                                return ranks
            else:
                self._res_evt.wait(timeout=0.1)


    # --------------------------------------------------------------------------
    #
    def _dealloc(self, task):
        '''
        deallocate task ranks
        '''

        uid   = task['uid']
        ranks = task['ranks']

        with self._res_lock:

            for rank in ranks:
                self._resources['cores'][rank] = FREE

            # signal available resources
            self._res_evt.set()
            self._prof.prof('unschedule_stop', uid=uid)

        # remove temporary information from task
        del task['rank']
        del task['ranks']


# ------------------------------------------------------------------------------
#
class _TaskPuller(mt.Thread):
    '''
    This class will pull tasks from the master, allocate suitable ranks for
    it's execution, and push the task to those ranks
    '''

    def __init__(self, worker_task_q_get, worker_result_q_put,
                       rank_task_q_put, event, resources, log, prof):

        super().__init__()

        self.daemon               = True
        self._worker_task_q_get   = worker_task_q_get
        self._worker_result_q_put = worker_result_q_put
        self._rank_task_q_put     = rank_task_q_put
        self._event               = event
        self._resources           = resources
        self._log                 = log
        self._prof                = prof
        self._ranks               = self._resources.ranks


    # --------------------------------------------------------------------------
    #
    def run(self):
        '''
        This callback gets tasks from the master, schedules resources for the
        tasks, and pushes them out to the respective ranks for execution.  If
        a task arrives for which no resources are available, the thread will
        block until such resources do become available.
        '''
        self._log.debug('init task puller 0 wtq_get:%s wrq_put:%s rtq_put:%s',
                        self._worker_task_q_get, self._worker_result_q_put,
                        self._rank_task_q_put)

        try:
            # register callback to receive tasks
            # connect to the master's task queue
            worker_task_q = ru.zmq.Getter('raptor_tasks',
                                          url=self._worker_task_q_get,
                                          log=self._log, prof=self._prof)

            # send tasks to worker ranks
            rank_task_q = ru.zmq.Putter('rank_tasks', url=self._rank_task_q_put,
                                        log=self._log, prof=self._prof)

            # also connect to the master's result queue to inform about errors
            worker_result_q = ru.zmq.Putter('raptor_results',
                                            url=self._worker_result_q_put,
                                            log=self._log, prof=self._prof)

            time.sleep(1)
            # setup is completed - signal main thread
            self._event.set()

            while True:

                tasks = None
                try:
                    tasks = worker_task_q.get_nowait(timeout=1)
                except:
                    self._log.exception('pull error')

                if not tasks:
                    continue

                tasks = ru.as_list(tasks)
                self._log.debug('wtq tasks: %s', len(tasks))

                # TODO: sort tasks by size
                for task in ru.as_list(tasks):

                    self._log.debug('wtq %s 0 - task pulled', task['uid'])

                    try:
                        task['ranks'] = self._resources._alloc(task)
                        self._prof.prof('advance', uid=task['uid'],
                                        state=AGENT_EXECUTING_PENDING)
                        for rank in task['ranks']:
                            task['rank'] = rank
                            self._log.debug('wtq %s 1 - task send to %d %s',
                                             task['uid'], rank, task['ranks'])
                            rank_task_q.put(task, qname=str(rank))

                    except Exception as e:
                        self._log.exception('failed to place task')
                        task['exception']        = repr(e)
                        task['exception_detail'] = \
                                             '\n'.join(ru.get_exception_trace())
                        worker_result_q.put(task)

        except:
            self._log.exception('task puller cb failed')


# --------------------------------------------------------------------------
#
class _ResultPusher(mt.Thread):
    '''
    This helper class will wait for result messages from ranks which completed
    the execution of a task.  It will collect results from all ranks which
    belong to that specific task and then send the results back to the master.
    '''

    def __init__(self, worker_result_q_put, rank_result_q_get, event,
                       resources, log, prof):

        super().__init__()

        self.daemon               = True
        self._worker_result_q_put = worker_result_q_put
        self._rank_result_q_get   = rank_result_q_get
        self._event               = event
        self._resources           = resources
        self._log                 = log
        self._prof                = prof


    # --------------------------------------------------------------------------
    #
    def _check_ranks(self, task):
        '''
        collect results of task ranks

        Returns `True` once all ranks are collected - the task then contains the
        collected results
        '''
        uid   = task['uid']
        ranks = task['description'].get('ranks', 1)

        if uid not in self._cache:
            self._cache[uid] = list()

        self._cache[uid].append(task)

        # do we have all ranks?
        if len(self._cache[uid]) < ranks:
            return False

        task['stdout']       = [t['stdout']       for t in self._cache[uid]]
        task['stderr']       = [t['stderr']       for t in self._cache[uid]]
        task['return_value'] = [t['return_value'] for t in self._cache[uid]]

        exit_codes           = [t['exit_code']    for t in self._cache[uid]]
        task['exit_code']    = sorted(list(set(exit_codes)))[-1]

        return True


    # --------------------------------------------------------------------------
    #
    def run(self):
        '''
        This thread pulls tasks from the master, schedules resources for the
        tasks, and pushes them out to the respective ranks for execution.  If
        a task arrives for which no resources are available, the thread will
        block until such resources do become available.
        '''

        try:
            self._log.debug('init result pusher wrq_put:%s rrq_get:%s',
                            self._worker_result_q_put, self._rank_result_q_get)

            # collect the results from all MPI ranks before returning
            self._cache = dict()

            # collect results from worker ranks
            rank_result_q = ru.zmq.Getter(channel='rank_results',
                                          url=self._rank_result_q_get,
                                          log=self._log,
                                          prof=self._prof)

            # collect results from worker ranks
            worker_result_q = ru.zmq.Putter(channel='raptor_results',
                                            url=self._worker_result_q_put,
                                            log=self._log, prof=self._prof)

            time.sleep(1)
            # signal success
            self._event.set()

            while True:

                # FIXME: use poller of callback
                tasks = ru.as_list(rank_result_q.get_nowait(timeout=100))

                for task in tasks:

                    self._log.debug('rrq %s %s [%s]', task['uid'],
                            task['rank'], task['ranks'])

                    # did all ranks complete?
                    if self._check_ranks(task):
                        self._resources._dealloc(task)
                        worker_result_q.put(task)


        except:
            self._log.exception('result pusher thread failed')


# ------------------------------------------------------------------------------
#
class MPIWorkerRank(mt.Thread):

    # --------------------------------------------------------------------------
    #
    def __init__(self, rank_task_q_get, rank_result_q_put,
                       mpi_info, event, log, prof, base):

        super().__init__()

        self.daemon             = True
        self._rank_task_q_get   = rank_task_q_get
        self._rank_result_q_put = rank_result_q_put

        self._world             = mpi_info['world']
        self._group             = mpi_info['group']
        self._rank              = mpi_info['rank' ]
        self._ranks             = mpi_info['ranks']

        self._event             = event
        self._log               = log
        self._prof              = prof
        self._base              = base
        self._sbox              = os.environ['RP_TASK_SANDBOX']


    # --------------------------------------------------------------------------
    #
    def run(self):

        try:
            self._log.debug('init worker [%d] [%d] rtq_get:%s rrq_put:%s',
                            self._rank, self._ranks,
                            self._rank_task_q_get, self._rank_result_q_put)

            # get tasks from rank 0
            rank_task_q = ru.zmq.Getter('rank_tasks', url=self._rank_task_q_get,
                                        log=self._log, prof=self._prof)
            # send results back to rank 0
            rank_result_q = ru.zmq.Putter('rank_results',
                                          url=self._rank_result_q_put,
                                          log=self._log,  prof=self._prof)

            # signal success
            self._event.set()

            # get task, run it, push results back
            while True:

                tasks = rank_task_q.get_nowait(qname=str(self._rank), timeout=100)

                if not tasks:
                    continue

                assert len(tasks) == 1

                task = tasks[0]
                uid  = task['uid']
                self._prof.prof('advance',    uid=uid, state=AGENT_EXECUTING)
                self._prof.prof('task_start', uid=uid)

                try:
                    # this should never happen
                    if self._rank not in task['ranks']:
                        raise RuntimeError('inconsistent rank info')

                    sbox = task['task_sandbox_path']
                    ru.rec_makedir(sbox)
                    os.chdir(sbox)

                    self._prof.prof('exec_start', uid=uid)
                    out, err, ret, val, exc = self._dispatch(task)
                    self._prof.prof('exec_stop', uid=uid)

                    task['stdout']           = out
                    task['stderr']           = err
                    task['exit_code']        = ret
                    task['return_value']     = val
                    task['exception']        = exc[0]
                    task['exception_detail'] = exc[1]

                except Exception as e:
                    task['stdout']           = ''
                    task['stderr']           = str(e)
                    task['exit_code']        = -1
                    task['return_value']     = None
                    task['exception']        = repr(e)
                    task['exception_detail'] = \
                                         '\n'.join(ru.get_exception_trace())
                    self._log.exception('recv err  %s  to  0' % (task['uid']))

                finally:
                    # send task back to rank 0
                    # FIXME: task_exec_stop
                    os.chdir(self._sbox)
                    self._prof.prof('unschedule_start', uid=uid)
                    rank_result_q.put(task)

                  # if task['uid'] == 'task.call_mpi.c.000000':
                  #     raise RuntimeError('oops')

        except:
            self._log.exception('work thread failed')
            self._base.stop(1)

        else:
            self._base.stop()


    # --------------------------------------------------------------------------
    #
    def _dispatch(self, task):

        task['description']['environment'].update(
              {'RP_TASK_ID'         : task['uid'],
               'RP_TASK_NAME'       : task.get('name'),
               'RP_TASK_SANDBOX'    : os.environ['RP_TASK_SANDBOX'],  # FIXME?
               'RP_PILOT_ID'        : os.environ['RP_PILOT_ID'],
               'RP_SESSION_ID'      : os.environ['RP_SESSION_ID'],
               'RP_RESOURCE'        : os.environ['RP_RESOURCE'],
               'RP_RESOURCE_SANDBOX': os.environ['RP_RESOURCE_SANDBOX'],
               'RP_SESSION_SANDBOX' : os.environ['RP_SESSION_SANDBOX'],
               'RP_PILOT_SANDBOX'   : os.environ['RP_PILOT_SANDBOX'],
               'RP_GTOD'            : os.environ['RP_GTOD'],
               'RP_PROF'            : os.environ['RP_PROF'],
               'RP_PROF_TGT'        : os.environ['RP_PROF_TGT'],
               'RP_RANKS'           : '1',  # dispatch_mpi will overwrite this
               'RP_RANK'            : '0',  # dispatch_mpi will overwrite this
               })

        if task['description']['ranks'] > 1:
            return self._dispatch_mpi(task)
        else:
            return self._dispatch_non_mpi(task)



    # --------------------------------------------------------------------------
    #
    def _dispatch_mpi(self, task):

        comm  = None
        group = None

        # create new communicator with all workers assigned to this task
        group = self._group.Incl(task['ranks'])
        comm  = self._world.Create_group(group)
        if not comm:
            out = None
            err = 'MPI setup failed'
            ret = 1
            val = None
            exc = (None, None)
            return out, err, ret, val, exc

        task['description']['environment']['RP_RANK']   = str(comm.rank)
        task['description']['environment']['RP_RANKS']  = str(comm.size)

        task['mpi_comm'] = comm

        try:
            return self._dispatch_non_mpi(task)

        finally:
            if 'mpi_comm' in task:
                del task['mpi_comm']

            # sub-communicator must always be destroyed
            if group: group.Free()
            if comm : comm.Free()


    # --------------------------------------------------------------------------
    #
    def _dispatch_non_mpi(self, task):

        # work on task
        mode       = task['description']['mode']
        dispatcher = self._base.get_dispatcher(mode)

        if not dispatcher:
            raise ValueError('no execution mode defined for %s' % mode)

        return dispatcher(task)


# ------------------------------------------------------------------------------
#
class MPIWorker(Worker):
    '''
    This worker manages a certain number of cores and gpus.  The master will
    start this worker by placing one rank per managed core (the GPUs are used
    dynamically).

    The first rank (rank 0) will manage the worker and for that purpose spawns
    two threads.  The first will pull tasks from the master's queue, and upon
    arrival will:

      - schedule incoming tasks over the available ranks
      - sent each target rank the required task startup info

    The second thread will collect the results from the tasks and send them back
    to the master.  The communication between rank 0 and the other ranks is
    through two ZMQ queues which are created and managed by rank 0.

    The main thread of rank 0 will function like the other threads: wait for
    task startup info and enact them.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, raptor_id: str):

        self._my_term = mt.Event()
        self._my_ret  = 0

        from mpi4py import MPI                                            # noqa

        self._world = MPI.COMM_WORLD
        self._group = self._world.Get_group()

        self._rank  = self._world.rank
        self._ranks = self._world.size

        if self._rank == 0: self._manager = True
        else              : self._manager = False

        # rank 0 is the manager rank and will register the worker with the
        # master and connect to the task and result queues
        super().__init__(manager=self._manager, rank=self._rank,
                         raptor_id=raptor_id)


        # rank 0 starts two ZMQ queues: one to send tasks to the worker ranks
        # (rank_task_q), and one to collect results from the ranks
        # (rank_result_q)
        info = None
        if self._manager:

            self._worker_task_q_get   = self._req_addr_get
            self._worker_result_q_put = self._res_addr_put

            self._rank_task_q   = ru.zmq.Queue(channel='rank_task_q')
            self._rank_result_q = ru.zmq.Queue(channel='rank_result_q')

            self._rank_task_q.start()
            self._rank_result_q.start()

            info = {'rank_task_q_put'  : str(self._rank_task_q.addr_put),
                    'rank_task_q_get'  : str(self._rank_task_q.addr_get),
                    'rank_result_q_put': str(self._rank_result_q.addr_put),
                    'rank_result_q_get': str(self._rank_result_q.addr_get)}

            # let channels settle
            time.sleep(1)

        # broadcast the queue endpoint addresses to all worker ranks
        info = self._world.bcast(info, root=0)

        self._rank_result_q_put = info['rank_result_q_put']
        self._rank_result_q_get = info['rank_result_q_get']
        self._rank_task_q_put   = info['rank_task_q_put']
        self._rank_task_q_get   = info['rank_task_q_get']


    # --------------------------------------------------------------------------
    #
    def get_rank_worker(self):
        '''Return the class type of the worker class which manages each worker
        rank.
        '''

        return MPIWorkerRank


    # --------------------------------------------------------------------------
    #
    def start(self):
        '''
        start this worker
        '''

        # all ranks run a worker thread
        # the worker should be started before the managers as the manager
        # contacts the workers with queue endpoint information
        worker_ok = mt.Event()
        worker    = self.get_rank_worker()

        mpi_info  = {'world': self._world,
                     'group': self._group,
                     'rank' : self._rank,
                     'ranks': self._ranks,
                     'rank0': self._manager}
        self._work_thread = worker(rank_task_q_get   = self._rank_task_q_get,
                                   rank_result_q_put = self._rank_result_q_put,
                                   mpi_info          = mpi_info,
                                   event             = worker_ok,
                                   log               = self._log,
                                   prof              = self._prof,
                                   base              = self)
        self._work_thread.start()
        worker_ok.wait(timeout=60)

        if not worker_ok.is_set():
            raise RuntimeError('failed to start worker thread')


        # the manager (rank 0) will start two threads - one to pull tasks from
        # the master, one to push results back to the master
        if self._manager:

            resources = _Resources(self._log,   self._prof, self._ranks)

            # rank 0 spawns manager threads
            pull_ok = mt.Event()
            push_ok = mt.Event()

            self._log.debug('wrq_put:%s wtq_get:%s',
                            self._worker_result_q_put, self._worker_task_q_get)

            self._pull = _TaskPuller(
                    worker_task_q_get   = self._worker_task_q_get,
                    worker_result_q_put = self._worker_result_q_put,
                    rank_task_q_put     = self._rank_task_q.addr_put,
                    event               = pull_ok,
                    resources           = resources,
                    log                 = self._log,
                    prof                = self._prof)

            self._push = _ResultPusher(
                    worker_result_q_put = self._worker_result_q_put,
                    rank_result_q_get   = self._rank_result_q.addr_get,
                    event               = push_ok,
                    resources           = resources,
                    log                 = self._log,
                    prof                = self._prof)

            self._pull.start()
            self._push.start()

            pull_ok.wait(timeout=60)
            push_ok.wait(timeout=60)

            if not pull_ok.is_set():
                raise RuntimeError('failed to start pull thread')

            if not push_ok.is_set():
                raise RuntimeError('failed to start push thread')


    # --------------------------------------------------------------------------
    #
    def stop(self, ret=0):
        '''
        terminate this worker
        '''

        self._log.debug('stop: set term signal')
        self._my_ret = ret
        self._my_term.set()


    # --------------------------------------------------------------------------
    #
    def join(self):
        '''
        block until this worker receives a termination signal
        '''

        # FIXME
        while not self._my_term.is_set():
            if self._my_term.wait(1):
                break

        self._log.debug('stop: term signal set - joined: %s', self._my_ret)

        if self._my_ret:
            raise RuntimeError('MPI worker failed with non-zero exit code')


    # --------------------------------------------------------------------------
    #
    def _call(self, task):
        '''
        We expect data to have a three entries: 'method' or 'function',
        containing the name of the member method or the name of a free function
        to call, `args`, an optional list of unnamed parameters, and `kwargs`,
        and optional dictionary of named parameters.
        '''

        data = task['data']

        if 'method' in data:
            to_call = getattr(self, data['method'], None)

        elif 'function' in data:
            names   = dict(list(globals().items()) + list(locals().items()))
            to_call = names.get(data['function'])

        else:
            raise ValueError('no method or function specified: %s' % data)

        if not to_call:
            raise ValueError('callable not found: %s' % data)

        args   = data.get('args',   [])
        kwargs = data.get('kwargs', {})

        bak_stdout = sys.stdout
        bak_stderr = sys.stderr

        strout = None
        strerr = None

        try:
            # redirect stdio to capture them during execution
            sys.stdout = strout = io.StringIO()
            sys.stderr = strerr = io.StringIO()

            val = to_call(*args, **kwargs)
            out = strout.getvalue()
            err = strerr.getvalue()
            exc = (None, None)
            ret = 0

        except Exception as e:
            self._log.exception('_call failed: %s' % (data))
            val = None
            out = strout.getvalue()
            err = strerr.getvalue() + ('\ncall failed: %s' % e)
            exc = (repr(e), '\n'.join(ru.get_exception_trace()))
            ret = 1

        finally:
            # restore stdio
            sys.stdout = bak_stdout
            sys.stderr = bak_stderr

        res = [task, out, err, ret, val, exc]

        return res


    # --------------------------------------------------------------------------
    #
    def hello_mpi(self, comm, msg, sleep=0):

        msg = '%s [%d/%d]' % (msg, comm.rank, comm.size)

        return self.hello(msg, sleep)


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import time

import threading     as mt

import radical.utils as ru

from ...  import states    as rps
from ...  import constants as rpc

from .base import AgentExecutingComponent


# ------------------------------------------------------------------------------
#
class Sleep(AgentExecutingComponent) :

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        AgentExecutingComponent.__init__ (self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        self._pwd = os.getcwd()

        self.register_input(rps.AGENT_EXECUTING_PENDING,
                            rpc.AGENT_EXECUTING_QUEUE, self.work)

        self.register_output(rps.AGENT_STAGING_OUTPUT_PENDING,
                             rpc.AGENT_STAGING_OUTPUT_QUEUE)

        self.register_publisher(rpc.AGENT_UNSCHEDULE_PUBSUB)

        self._terminate  = mt.Event()
        self._tasks_lock = mt.RLock()
        self._tasks      = list()
        self._delay      = 1.0

        self._watcher = mt.Thread(target=self._collect)
        self._watcher.daemon = True
        self._watcher.start()


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        self._terminate.set()
        self._watcher.join()


    # --------------------------------------------------------------------------
    #
    def work(self, tasks):

        self.advance_tasks(tasks, rps.AGENT_EXECUTING, publish=True, push=False)

        for task in tasks:

            try:
                self._prof.prof('task_start', uid=task['uid'])
                self._handle_task(task)

            except Exception as e:
                self._log.exception("error running Task")
                task['exception']        = repr(e)
                task['exception_detail'] = '\n'.join(ru.get_exception_trace())

                # can't rely on the executor base to free the task resources
                self._prof.prof('unschedule_start', uid=task['uid'])
                self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)

                self.advance_tasks(task, rps.FAILED, publish=True, push=False)

        with self._tasks_lock:
            self._tasks.extend(tasks)


    # --------------------------------------------------------------------------
    #
    def cancel_task(self, uid):

        raise NotImplementedError('no cancellation support in sleep executor')


    # --------------------------------------------------------------------------
    #
    def _handle_task(self, task):

        now = time.time()

        # assert t['description']['executable'].endswith('sleep')
        task['to_finish'] = now + float(task['description']['arguments'][0])

        uid = task['uid']
        self._prof.prof('task_run_start', uid=uid)
        self._prof.prof('task_run_ok',    uid=uid)
        self._prof.prof('launch_start',   uid=uid)
        self._prof.prof('exec_start',     uid=uid)
        self._prof.prof('rank_start',      uid=uid)


    # --------------------------------------------------------------------------
    #
    def _collect(self):

        while not self._terminate.is_set():

            to_finish   = list()
            to_continue = list()
            now         = time.time()

            with self._tasks_lock:

                for task in self._tasks:
                    if task['deadline'] <= now: to_finish.append(task)
                    else                      : to_continue.append(task)

                self._tasks = to_continue

            if not to_finish:
                time.sleep(self._delay)
                continue

            for task in to_finish:
                uid = task['uid']
                task['target_state'] = 'DONE'

                self._prof.prof('rank_stop',        uid=uid)
                self._prof.prof('exec_stop',        uid=uid)
                self._prof.prof('launch_stop',      uid=uid)
                self._prof.prof('task_run_stop',    uid=uid)
                self._prof.prof('unschedule_start', uid=uid)

            self._log.debug('collected                : %d', len(to_finish))

            self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, to_finish)
            self.advance_tasks(to_finish, rps.AGENT_STAGING_OUTPUT_PENDING,
                                          publish=True, push=True)


    # --------------------------------------------------------------------------
    #
    def control_cb(self, topic, msg):

        self._log.info('control_cb [%s]: %s', topic, msg)

        cmd = msg['cmd']

        # FIXME RPC: already handled in the component base class
        if cmd == 'cancel_tasks':

            # FIXME: clarify how to cancel tasks
            pass


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2023-2024, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os

import radical.utils as ru

from .continuous import Continuous


# ------------------------------------------------------------------------------
#
class ContinuousReconfig(Continuous):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        super().__init__(cfg, session)

        # file with updated resource requirements (ranks, cores_per_rank)
        self._reconfig_src = self._cfg.reconfig_src
        self._task_reqs    = {}

    # --------------------------------------------------------------------------
    #
    def work(self, tasks):

        # check possibly updated task requirements
        if self._reconfig_src and os.path.isfile(self._reconfig_src):
            task_reqs = ru.read_json(self._reconfig_src)
            if self._task_reqs != task_reqs:
                self._log.debug('set new task reqs: %s', task_reqs)
                self._task_reqs = task_reqs

        if not self._task_reqs:
            return super().work(tasks)

        tasks_updated = ru.as_list(tasks)
        for attr in ['ranks', 'cores_per_rank']:
            v = self._task_reqs.get(attr)
            if v is not None:
                v = int(v)
                for task in tasks_updated:
                    task['description'][attr] = v

        super().work(tasks_updated)


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2020-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import collections
import math
import os
import signal
import time

import subprocess    as mp
import threading     as mt

import radical.utils as ru

from .base import LaunchMethod

PTL_MAX_MSG_SIZE = 1024 * 1024 * 1024 * 1

DVM_URI_FILE_TPL   = '%(base_path)s/prrte.%(dvm_id)04d.uri'
DVM_HOSTS_FILE_TPL = '%(base_path)s/prrte.%(dvm_id)04d.hosts'


# ------------------------------------------------------------------------------
#
class PRTE(LaunchMethod):
    """
    PMIx Reference Run Time Environment v2
    (https://github.com/openpmix/prrte)
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, lm_cfg, rm_info, log, prof):

        self._verbose = bool(os.environ.get('RADICAL_PILOT_PRUN_VERBOSE'))

        LaunchMethod.__init__(self, name, lm_cfg, rm_info, log, prof)

    # --------------------------------------------------------------------------
    #
    def __del__(self):

        self._terminate()

    # --------------------------------------------------------------------------
    #
    def _configure(self):

        prte_cmd = ru.which('prte')
        if not prte_cmd:
            raise Exception('prte command not found')

        prte_prefix = os.environ.get('UMS_OMPIX_PRRTE_DIR')

        prte_info = {}  # get OpenRTE/PRTE version
        out = ru.sh_callout('prte_info | grep "RTE"', shell=True)[0]
        for line in out.split('\n'):

            line = line.strip()

            if not prte_info.get('name'):
                if   'Open RTE' in line: prte_info['name'] = 'Open RTE'
                elif 'PRTE'     in line: prte_info['name'] = 'PRTE'

            if 'RTE:' in line:
                prte_info['version'] = line.split(':')[1].strip()
            elif 'RTE repo revision:' in line:
                prte_info['version_detail'] = line.split(':')[1].strip()

        if prte_info.get('name'):
            self._log.info('version of %s: %s [%s]',
                           prte_info['name'],
                           prte_info.get('version'),
                           prte_info.get('version_detail'))
        else:
            self._log.info('version of OpenRTE/PRTE not determined')

        # get `dvm_count` from resource config (LM config section)
        # FIXME: another option is to introduce `nodes_per_dvm` in LM config
        #        dvm_count = math.ceil(len(rm_info.node_list) / nodes_per_dvm)
        dvm_count     = self._lm_cfg.get('dvm_count') or 1
        nodes_per_dvm = math.ceil(len(self._rm_info.node_list) / dvm_count)

        # additional info to form prrte related files (uri-/hosts-file)
        dvm_file_info = {'base_path': os.getcwd()}

        # ----------------------------------------------------------------------
        def _watch_dvm(dvm_process, dvm_id, dvm_ready_flag):

            self._log.info('prte-%s watcher started', dvm_id)

            retval = dvm_process.poll()
            while retval is None:
                line = dvm_process.stdout.readline().strip()
                if line:
                    self._log.debug('prte-%s output: %s', dvm_id, line)

                    # final confirmation that DVM has started successfully
                    if not dvm_ready_flag.is_set() and 'DVM ready' in str(line):
                        dvm_ready_flag.set()
                        self._log.info('prte-%s is ready', dvm_id)
                        self._prof.prof(event='dvm_ready',
                                        uid=self._lm_cfg['pid'],
                                        msg='dvm_id=%s' % dvm_id)

                else:
                    time.sleep(1.)

            if retval != 0:
                # send a kill signal to the main thread.
                # We know that Python and threading are likely not to play well
                # with signals - but this is an exceptional case, and not part
                # of the standard termination sequence. If the signal is
                # swallowed, the next prun-call will trigger termination anyway.
                os.kill(os.getpid(), signal.SIGKILL)
                raise RuntimeError('PRTE DVM died (dvm_id=%s)' % dvm_id)

            self._log.info('prte-%s watcher stopped', dvm_id)

        # ----------------------------------------------------------------------
        def _start_dvm(dvm_id, dvm_size, dvm_ready_flag):

            file_info = dict(dvm_file_info)
            file_info['dvm_id'] = dvm_id

            prte  = '%s' % prte_cmd
            if prte_prefix: prte += ' --prefix %s' % prte_prefix
            prte += ' --report-uri %s' % DVM_URI_FILE_TPL   % file_info
            prte += ' --hostfile %s'   % DVM_HOSTS_FILE_TPL % file_info

            # large tasks imply large message sizes, we need to account for that
            # FIXME: need to derive the message size from DVM size - smaller
            #        DVMs will never need large messages, as they can't run
            #        large tasks
            prte += ' --pmixmca ptl_base_max_msg_size %d'  % PTL_MAX_MSG_SIZE
            prte += ' --prtemca routed_radix %d'           % dvm_size

            # ensure 1 ssh per dvm (all daemons are directly connected to prte)
            prte += ' --prtemca plm_rsh_num_concurrent %d' % dvm_size

            # ssh connection limit (avoid 64 node limit)
            #   --prtemca plm_rsh_no_tree_spawn 1'

            # select a set of hardware threads that should be used for
            # processing the following option should set hwthread IDs:
            #   --prtemca hwloc_default_cpu_list "0-83,88-171"

            # Use (g)stdbuf to disable buffering.  We need this to get the
            # "DVM ready" message to ensure DVM startup completion
            #
            # The command seems to be generally available on Cray's,
            # if not, we can code some home-cooked pty stuff (TODO)
            stdbuf_cmd = ru.which(['stdbuf', 'gstdbuf'])
            if not stdbuf_cmd:
                raise Exception('(g)stdbuf command not found')
            stdbuf_arg = '-oL'

            # base command = (g)stdbuf <args> + prte <args>
            cmd = '%s %s %s ' % (stdbuf_cmd, stdbuf_arg, prte)

            # additional (debug) arguments to prte
            if self._verbose:
                cmd += ' '.join(['--prtemca plm_base_verbose 5'])

            cmd = cmd.strip()

            self._log.info('start prte-%s on %d nodes [%s]',
                           dvm_id, dvm_size, cmd)
            self._prof.prof(event='dvm_start',
                            uid=self._lm_cfg['pid'],
                            msg='dvm_id=%s' % dvm_id)

            dvm_process = mp.Popen(cmd.split(),
                                   stdout=mp.PIPE,
                                   stderr=mp.STDOUT)
            dvm_watcher = mt.Thread(target=_watch_dvm,
                                    args=(dvm_process, dvm_id, dvm_ready_flag),
                                    daemon=True)
            dvm_watcher.start()

            dvm_uri = None
            for _ in range(30):
                time.sleep(.5)

                try:
                    with ru.ru_open(DVM_URI_FILE_TPL % file_info, 'r') as fin:
                        for line in fin.readlines():
                            if '://' in line:
                                dvm_uri = line.strip()
                                break
                except Exception as e:
                    self._log.debug('DVM URI file missing: %s...', str(e)[:24])

                if dvm_uri:
                    break

            if not dvm_uri:
                raise Exception('DVM URI not found')

            self._log.info('prte-%s startup successful: [%s]', dvm_id, dvm_uri)
            self._prof.prof(event='dvm_uri',
                            uid=self._lm_cfg['pid'],
                            msg='dvm_id=%s' % dvm_id)

            return dvm_uri

        # ----------------------------------------------------------------------

        # format: {<id>: {'nodes': [...], 'dvm_uri': <uri>}}
        dvm_list = {}

        # go through every dvm instance
        for _dvm_id in range(dvm_count):

            node_list = self._rm_info.node_list[_dvm_id      * nodes_per_dvm:
                                               (_dvm_id + 1) * nodes_per_dvm]
            dvm_file_info.update({'dvm_id': _dvm_id})
            # write hosts file
            with ru.ru_open(DVM_HOSTS_FILE_TPL % dvm_file_info, 'w') as fout:
                for node in node_list:
                    fout.write('%s slots=%d\n' % (node['node_name'],
                                                  self._rm_info.cores_per_node))

            _dvm_size  = len(node_list)
            _dvm_ready = mt.Event()
            _dvm_uri   = _start_dvm(_dvm_id, _dvm_size, _dvm_ready)

            dvm_list[str(_dvm_id)] = {
                'nodes'  : [node['node_id'] for node in node_list],
                'dvm_uri': _dvm_uri}

            # extra time to confirm that "DVM ready" was just delayed
            if _dvm_ready.wait(timeout=15.):
                continue

            self._log.info('prte-%s to be re-started', _dvm_id)

            # terminate DVM (if it is still running)
            try   : ru.sh_callout('pterm --dvm-uri "%s"' % _dvm_uri)
            except: pass

            # re-start DVM
            dvm_list[str(_dvm_id)]['dvm_uri'] = \
                _start_dvm(_dvm_id, _dvm_size, _dvm_ready)

            # FIXME: with the current approach there is only one attempt to
            #        restart DVM(s). If a failure during the start process
            #        will keep appears then need to consider re-assignment
            #        of nodes from failed DVM(s) + add time for re-started
            #        DVM to stabilize: `time.sleep(10.)`

        lm_details = {'dvm_list'    : dvm_list,
                      'version_info': prte_info}

        return lm_details


    # --------------------------------------------------------------------------
    #
    def get_partitions(self):

        partitions = dict()

        for k, v in self._details['dvm_list'].items():
            partitions[str(k)] = v['nodes']

        return partitions


    # --------------------------------------------------------------------------
    #
    def _terminate(self):
        """
        This hook is symmetric to the hook `rm_config_hook`, it is called
        during shutdown sequence, for the sake of freeing allocated resources.
        """
        cmd = ru.which('pterm')
        if not cmd:
            raise Exception('termination command not found')

        for p_id, p_data in self._details.get('dvm_list', {}).items():
            dvm_uri = p_data['dvm_uri']
            try:
                self._log.info('terminating prte-%s (%s)', p_id, dvm_uri)
                _, err, _ = ru.sh_callout('%s --dvm-uri "%s"' % (cmd, dvm_uri))
                self._log.debug('termination state: %s', err.strip('\n') or '-')
                self._prof.prof(event='dvm_stop', uid=self._lm_cfg['pid'],
                                msg='dvm_id=%s' % p_id)

            except Exception as e:
                # use the same event name as for runtime failures - those are
                # not distinguishable at the moment from termination failures
                self._log.debug('prte-%s termination failed (%s)', p_id, dvm_uri)
                self._prof.prof(event='dvm_fail', uid=self._lm_cfg['pid'],
                                msg='dvm_id=%s | error: %s' % (p_id, e))

    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, env, env_sh):

        lm_info = {'env'    : env,
                   'env_sh' : env_sh,
                   'command': ru.which('prun'),
                   'details': self._configure()}

        return lm_info

    # --------------------------------------------------------------------------
    #
    def _init_from_info(self, lm_info):

        self._env     = lm_info['env']
        self._env_sh  = lm_info['env_sh']
        self._command = lm_info['command']
        self._details = lm_info['details']

        assert self._command

    # --------------------------------------------------------------------------
    #
    def finalize(self):

        self._terminate()

    # --------------------------------------------------------------------------
    #
    def can_launch(self, task):

        if not task['description']['executable']:
            return False, 'no executable'

        return True, ''


    # --------------------------------------------------------------------------
    #
    def get_launcher_env(self):

        return ['. $RP_PILOT_SANDBOX/%s' % self._env_sh]

    # --------------------------------------------------------------------------
    #
    def get_launch_cmds(self, task, exec_path):

        # latest implementation of PRRTE is able to handle bulk submissions
        # of prun-commands, but in case errors will come out then the delay
        # should be set: `time.sleep(.1)`

        slots     = task['slots']
        td        = task['description']

        n_procs   = td['ranks']
        n_threads = td['cores_per_rank']

        if not self._details.get('dvm_list'):
            raise RuntimeError('details with dvm_list not set (%s)' % self.name)

        dvm_list  = self._details['dvm_list']
        # `partition_id` should be set in a scheduler
        dvm_id    = slots.get('partition_id') or list(dvm_list.keys())[0]
        dvm_uri   = '--dvm-uri "%s"' % dvm_list[dvm_id]['dvm_uri']

        flags  = '--np %d '                                   % n_procs
        flags += '--map-by node:HWTCPUS:PE=%d:OVERSUBSCRIBE ' % n_threads
        flags += '--bind-to hwthread:overload-allowed'

        if self._verbose:
            flags += ':REPORT'

        if 'ranks' not in slots:
            # this task is unscheduled - we leave it to PRRTE/PMI-X
            # to correctly place the task
            pass
        else:
            ranks = collections.defaultdict(int)
            for rank in slots['ranks']:
                ranks[rank['node_name']] += 1
            flags += ' --host ' + ','.join(['%s:%s' % x for x in ranks.items()])

        flags += ' --pmixmca ptl_base_max_msg_size %d' % PTL_MAX_MSG_SIZE
        flags += ' --verbose'  # needed to get prte profile events

        cmd = '%s %s %s %s' % (self._command, dvm_uri, flags, exec_path)

        return cmd.strip()

    # --------------------------------------------------------------------------
    #
    def get_rank_cmd(self):

        return 'export RP_RANK=$PMIX_RANK\n'


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import sys
import time
import radical.pilot as rp

# READ: The RADICAL-Pilot documentation:
#   https://radicalpilot.readthedocs.io/en/stable/
#
# Try running this example with RADICAL_PILOT_VERBOSE=debug set if
# you want to see what happens behind the scenes!


# ------------------------------------------------------------------------------
#
def pilot_state_cb (pilot, state):
    """ this callback is invoked on all pilot state changes """

    print("[Callback]: Pilot '%s' state: %s." % (pilot.uid, state))


# ------------------------------------------------------------------------------
#
def task_state_cb (task, state):
    """ this callback is invoked on all task state changes """

    print("[Callback]: Task  '%s' state: %s." % (task.uid, state))


# ------------------------------------------------------------------------------
#
def wait_queue_size_cb(tmgr, wait_queue_size):
    """
    this callback is called when the size of the task managers wait_queue
    changes.
    """
    print("[Callback]: TaskManager  '%s' wait_queue_size changed to %s."
        % (tmgr.uid, wait_queue_size))

    pilots = tmgr.get_pilots ()
    for pilot in pilots:
        print("pilot %s: %s" % (pilot.uid, pilot.state))

    if wait_queue_size == 0:
        for pilot in pilots:
            if pilot.state in [rp.PENDING_LAUNCH,
                                rp.LAUNCHING     ,
                                rp.PENDING_ACTIVE]:
                print("cancel pilot %s" % pilot.uid)
                tmgr.remove_pilot (pilot.uid)
                pilot.cancel ()


# ------------------------------------------------------------------------------
#
if __name__ == "__main__":

    # we can optionally pass session name to RP
    if len(sys.argv) > 1:
        session_name = sys.argv[1]
    else:
        session_name = None

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session(uid=session_name)
    print("session id: %s" % session.uid)

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        # Add a Pilot Manager. Pilot managers manage one or more Pilots.
        pmgr = rp.PilotManager (session=session)

        # Register our callback with the PilotManager. This callback will get
        # called every time any of the pilots managed by the PilotManager
        # change their state.
        pmgr.register_callback (pilot_state_cb)

        # Define a 4-core local pilot that runs for 10 minutes and cleans up
        # after itself.

        pdesc1 = rp.PilotDescription()
        pdesc1.resource = "local.localhost"
        pdesc1.runtime  = 10  # minutes
        pdesc1.cores    =  2

        pdesc2 = rp.PilotDescription()
        pdesc2.resource = "local.localhost"
        pdesc2.runtime  = 10  # minutes
        pdesc2.cores    =  2

        # Launch the pilots
        pilots = pmgr.submit_pilots([pdesc1, pdesc2])

        # wait for them to become active
        pmgr.wait_pilots (state=[rp.PMGR_ACTIVE, rp.DONE, rp.FAILED])


        # Combine the Pilot, the Tasks and a scheduler via
        # a TaskManager object.
        tmgr = rp.TaskManager (session   = session,
                               scheduler = rp.SCHEDULER_BACKFILLING)

        # Register our callback with the TaskManager. This callback will get
        # called every time any of the tasks managed by the TaskManager
        # change their state.
        tmgr.register_callback (task_state_cb, rp.TASK_STATE)

        # Register also a callback which tells us when all tasks have been
        # assigned to pilots
        tmgr.register_callback(wait_queue_size_cb, rp.WAIT_QUEUE_SIZE)


        # Add the previously created Pilot to the TaskManager.
        tmgr.add_pilots (pilots)

        # Create a workload of restartable Tasks (tasks).
        tds = []
        for task_count in range(0, 32):
            td = rp.TaskDescription()
            td.executable    = "/bin/sleep"
            td.arguments     = ["10"]
            td.restartable   = True

            tds.append(td)

        # Submit the previously created Task descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning Tasks to the Pilots.
        tasks = tmgr.submit_tasks(tds)

        # the pilots have a total of 4 cores, and run for 10 min.  A Task needs about
        # 10 seconds, so we can handle about 24 tasks per minute, and need a total
        # of about 3 minutes.  We now wait for 60 seconds, and then cancel the first
        # pilot.  The 2 tasks currently running on that pilot will fail, and
        # maybe 2 more which are being pre-fetched into the pilot at that stage
        # - all others should get rescheduled to the other pilot.
        time.sleep(60)
        pilots[0].wait(state=rp.PMGR_ACTIVE)
        pilots[0].cancel()

        # Wait for all tasks to reach a terminal state (DONE or FAILED).
        tmgr.wait_tasks()

        print('tasks all completed')
        print('----------------------------------------------------------------------')

        for task in tasks:
            task.wait()

        for task in tasks:
            print("* Task %s state: %s, exit code: %s"
                  % (task.uid, task.state, task.exit_code))

    except Exception as e:
        # Something unexpected happened in the pilot code above
        print("caught Exception: %s" % e)
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        print("need to exit now: %s" % e)

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.
        print("closing session")
        session.close ()

        # the above is equivalent to
        #
        #   session.close (terminate=True)
        #
        # it will thus both clean out the session's database record, and kill
        # all remaining pilots (none in our example).

# -------------------------------------------------------------------------------



#!/usr/bin/env python3

# Changing the ratio among types of tasks at runtime.
# Algorithmically:
# - Start MD simulation tasks, use all the available resources
# - upon termination of an MD sim task:
#   - if the aggregation threshold is reached, kill a sim task and
#     launch an Aggregation task
#   - else, launch a new sim task
# - upon termination of an Aggregation task, launch a ML training task (possibly
#   killing some of the sim tasks if it requires more resource)
# - upon termination of an ML training task:
#   - if learning threshold is reached, launch an Agent task;
#   - else, launch a sim task
# - Upon termination of an Agent task, kill all the tasks and goto i.


import os
import time
import random
import signal
import threading as mt

import radical.pilot as rp
import radical.utils as ru


# ------------------------------------------------------------------------------
#
class DDMD(object):

    # define task types (used as prefix on task-uid)
    TASK_MD_SIM    = 'md_sim'
    TASK_AGGREGATE = 'aggregate'
    TASK_ML_TRAIN  = 'ml_train'
    TASK_AGENT     = 'agent'

    TASK_TYPES     = [TASK_MD_SIM, TASK_AGGREGATE, TASK_ML_TRAIN, TASK_AGENT]

    # keep track of core usage
    cores_used     = 0

    # --------------------------------------------------------------------------
    #
    def __init__(self):

        # control flow table
        self._protocol = {self.TASK_MD_SIM   : self._control_md_sim,
                          self.TASK_AGGREGATE: self._control_aggregate,
                          self.TASK_ML_TRAIN : self._control_ml_train,
                          self.TASK_AGENT    : self._control_agent}

        self._glyphs   = {self.TASK_MD_SIM   : '#',
                          self.TASK_AGGREGATE: '+',
                          self.TASK_ML_TRAIN : '=',
                          self.TASK_AGENT    : '*'}

        # bookkeeping
        self._aggregated     =  0
        self._aggregated_max = 64  # aggregation threshold

        self._trained        =  0
        self._trained_max    = 16  # training threshold

        self._cores          = 16  # available resources
        self._cores_used     =  0

        self._lock           = mt.RLock()
        self._tasks          = {ttype: dict() for ttype in self.TASK_TYPES}
        self._final_tasks    = list()

        # silence RP reporter, use own
        os.environ['RADICAL_REPORT'] = 'false'
        self._rep = ru.Reporter('ddmd')
        self._rep.title('DDMD')

        # RP setup
        self._session = rp.Session()
        self._pmgr    = rp.PilotManager(session=self._session)
        self._tmgr    = rp.TaskManager(session=self._session)

        pdesc = rp.PilotDescription({'resource': 'local.localhost',
                                     'runtime' : 30,
                                     'cores'   : self._cores})
        self._pilot = self._pmgr.submit_pilots(pdesc)

        self._tmgr.add_pilots(self._pilot)
        self._tmgr.register_callback(self._checked_state_cb)


    # --------------------------------------------------------------------------
    #
    def __del__(self):

        self.close()


    # --------------------------------------------------------------------------
    #
    def close(self):

        if self._session is not None:
            self._session.close()
            self._session = None


    # --------------------------------------------------------------------------
    #
    def dump(self, task=None, msg=''):
        '''
        dump a representation of current task set to stdout
        '''

        # this assumes one core per task

        self._rep.plain('<<|')

        idle = self._cores

        n     = len(self._tasks[self.TASK_MD_SIM])
        idle -= n
        self._rep.ok('%s' % self._glyphs[self.TASK_MD_SIM] * n)

        n     = len(self._tasks[self.TASK_AGGREGATE])
        idle -= n
        self._rep.warn('%s' % self._glyphs[self.TASK_AGGREGATE] * n)

        n     = len(self._tasks[self.TASK_ML_TRAIN])
        idle -= n
        self._rep.error('%s' % self._glyphs[self.TASK_ML_TRAIN] * n)

        n     = len(self._tasks[self.TASK_AGENT])
        idle -= n
        self._rep.info('%s' % self._glyphs[self.TASK_AGENT] * n)

        self._rep.plain('%s' % '-' * idle +
                        '| %4d [%4d]' % (self._cores_used, self._cores))

        if task and msg:
            self._rep.plain(' %-15s: %s\n' % (task.uid, msg))
        else:
            if task:
                msg = task
            self._rep.plain(' %-15s: %s\n' % (' ', msg))


    # --------------------------------------------------------------------------
    #
    def start(self):
        '''
        submit initial set of MD similation tasks
        '''

        self.dump('submit MD simulations')

        # reset bookkeeping
        self._aggregated = 0
        self._trained    = 0
        self._cores_used = 0
        self._tasks      = {ttype: dict() for ttype in self._protocol}

        # run initial batch of MD_SIM tasks (assume one core per task)
        self._submit_task(self.TASK_MD_SIM, n=self._cores)

        self.dump('started %s md sims' % self._cores)



    # --------------------------------------------------------------------------
    #
    def stop(self):

        os.kill(os.getpid(), signal.SIGKILL)
        os.kill(os.getpid(), signal.SIGTERM)


    # --------------------------------------------------------------------------
    #
    def _get_ttype(self, uid):
        '''
        get task type from task uid
        '''

        ttype = uid.split('.')[0]

        assert ttype in self.TASK_TYPES, 'unknown task type: %s' % uid
        return ttype


    # --------------------------------------------------------------------------
    #
    def _submit_task(self, ttype, n=1):
        '''
        submit 'n' new tasks of specified type

        NOTE: all tasks are uniform for now: they use a single core and sleep
              for a random number (0..3) of seconds.
        '''

        with self._lock:

            tds   = list()
            for _ in range(n):
                tds.append(rp.TaskDescription({
                         'uid'          : ru.generate_id(ttype),
                         'cpu_processes': 1,
                         'executable'   : '/bin/sh',
                         'arguments'    : ['-c', 'sleep %s; echo %s' %
                             (int(random.randint(0,30) / 10),
                              int(random.randint(0,10) /  1))]}))

            tasks  = self._tmgr.submit_tasks(tds)

            for task in tasks:
                self._register_task(task)


    # --------------------------------------------------------------------------
    #
    def _cancel_tasks(self, uids):
        '''
        cancel tasks with the given uids, and unregister them
        '''

        uids = ru.as_list(uids)

        # FIXME: does not work
        self._tmgr.cancel_tasks(uids)

        for uid in uids:
            ttype = self._get_ttype(uid)
            task  = self._tasks[ttype][uid]
            self.dump(task, 'cancel [%s]' % task.state)

            self._unregister_task(task)


    # --------------------------------------------------------------------------
    #
    def _register_task(self, task):
        '''
        add task to bookkeeping
        '''

        with self._lock:
            ttype = self._get_ttype(task.uid)
            self._tasks[ttype][task.uid] = task

            cores = task.description['cpu_processes'] \
                  * task.description['cpu_threads']
            self._cores_used += cores


    # --------------------------------------------------------------------------
    #
    def _unregister_task(self, task):
        '''
        remove completed task from bookkeeping
        '''

        with self._lock:

            ttype = self._get_ttype(task.uid)

            if task.uid not in self._tasks[ttype]:
                return

            # removed tasks dont consume cores
            cores = task.description['cpu_processes'] \
                  * task.description['cpu_threads']
            self._cores_used -= cores

            # remove task from bookkeeping
            self._final_tasks.append(task.uid)
            del self._tasks[ttype][task.uid]


    # --------------------------------------------------------------------------
    #
    def _state_cb(self, task, state):
        '''
        act on task state changes according to our protocol
        '''

        try:
            return self._checked_state_cb(task, state)
        except Exception as e:
            self._rep.error('\n\n---------\nexception caught: %s\n\n' % repr(e))
            self.stop()


    # --------------------------------------------------------------------------
    #
    def _checked_state_cb(self, task, state):

        # this cb will react on task state changes.  Specifically it will watch
        # out for task completion notification and react on them, depending on
        # the task type.

      # if state in [rp.TMGR_SCHEDULING] + rp.FINAL:
      #     self.dump(task, ' -> %s' % task.state)

        # ignore all non-final state transitions
        if state not in rp.FINAL:
            return

        # ignore tasks which were already
        if task.uid in self._final_tasks:
            return

        # lock bookkeeping
        with self._lock:

            # raise alarm on failing tasks (but continue anyway)
            if state == rp.FAILED:
                self._rep.error('task %s failed: %s' % (task.uid, task.stderr))
                self.stop()

            # control flow depends on ttype
            ttype  = self._get_ttype(task.uid)
            action = self._protocol[ttype]
            if not action:
                self._rep.exit('no action found for task %s' % task.uid)
            action(task)

            # remove final task from bookkeeping
            self._unregister_task(task)


    # --------------------------------------------------------------------------
    #
    def _control_md_sim(self, task):
        '''
        react on completed MD simulation task
        '''

        # - upon termination of an MD sim task:
        #   - if the aggregation threshold is reached,
        #     - launch an Aggregation task
        #   - else
        #     - launch a new sim task
        try:
            self._aggregated += int(task.stdout)
        except:
            pass

        if self._aggregated >= self._aggregated_max:
            self._aggregated = 0
            self.dump(task, 'completed, aggregation full - start aggregate')
            self._submit_task(self.TASK_AGGREGATE)
        else:
            self.dump(task, 'completed, aggregation low  - start md sim')
            self._submit_task(self.TASK_MD_SIM)


    # --------------------------------------------------------------------------
    #
    def _control_aggregate(self, task):
        '''
        react on completed aggregation task
        '''

        # - upon termination of an Aggregation task, launch a ML training task
        #   possibly killing some of the sim tasks if it requires more resources

        sim_uid = None
        while self._cores_used >= self._cores:

            if not self._tasks[self.TASK_MD_SIM]:
                # we can't free any resources - continue to submit aggregator
                break

            # kill a sim_task
            # FIXME: I think this is wrong: *this* aggregate task just finished
            #        anyway and thus there should be space to start a new sim
            sim_uid = random.choice(list(self._tasks[self.TASK_MD_SIM].keys()))
            self._cancel_tasks(sim_uid)
            break

        # submit training task
        self.dump(task, 'completed, cancel sim %s - start ml train ' % sim_uid)
        self._submit_task(self.TASK_ML_TRAIN)


    # --------------------------------------------------------------------------
    #
    def _control_ml_train(self, task):
        '''
        react on completed ML training task
        '''
        # - upon termination of an ML training task:
        #   - if learning threshold is reached
        #     - launch an Agent task;
        #   - else
        #     - launch a sim task

        self._trained += int(task.stdout)
        if self._trained >= self._trained_max:
            self.dump(task, 'completed, training complete - start agent ')
            self._submit_task(self.TASK_AGENT)
        else:
            self.dump(task, 'completed, training incomplete - start md sim ')
            self._submit_task(self.TASK_MD_SIM)


    # --------------------------------------------------------------------------
    #
    def _control_agent(self, task):
        '''
        react on completed agent task
        '''
        # - Upon termination of an Agent task, kill all the tasks and goto i.

        to_cancel = list()
        for ttype in self._tasks:
            to_cancel += list(self._tasks[ttype].keys())

        self.dump(task, 'completed, cancel all & restart')
        self._cancel_tasks(to_cancel)

        # restart execution
        self.start()


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    ddmd = DDMD()

    try:
        ddmd.start()

        while True:
          # ddmd.dump()
            time.sleep(1)

    finally:
        ddmd.close()


# ------------------------------------------------------------------------------

